/*
** Daedalus (Version 2.5) File: color.cpp
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains general graphics routines, that operate on a color
** bitmap, unrelated to Mazes.
**
** Old last code change: 6/29/1990.
** Last code change: 10/10/2013.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "util.h"
#include "graphics.h"
#include "color.h"


#define cchAscii 16
CONST char rgchAscii[cchAscii + 1] = "#MN@O$7%?*;:+,. ";

// Color RGB values and names.

CONST KV rgkv[cColor] = {
  kvBlack,  kvMaroon,  kvDkGreen, kvMaize,
  kvDkBlue, kvPurple,  kvDkCyan,  kvLtGray,
  kvGray,   kvRed,     kvGreen,   kvYellow,
  kvBlue,   kvMagenta, kvCyan,    kvWhite,

  kvMaize,  kvDkBlue,  kvPurple,  kvDkCyan,
  kvLtGray, kvGray,    kvGreen,   kvMagenta,
  kvCyan,   kvDkGray,  kvOrange,  kvPink,
  kvBrown};

CONST char *rgszColor[cColor] = {
  "Black",  "Maroon",  "DkGreen", "Maize",
  "DkBlue", "Purple",  "DkCyan",  "LtGray",
  "Gray",   "Red",     "Green",   "Yellow",
  "Blue",   "Magenta", "Cyan",    "White",

  "Olive",  "Navy",    "Violet",  "Teal",
  "Silver", "Grey",    "Lime",    "Fuchsia",
  "Aqua",   "DkGray",  "Orange",  "Pink",
  "Brown"};

CS cs = {fTrue, 0, nDegMax, kvBlack, kvRed, crpRing, fFalse, 299587114L,
  fFalse};


/*
******************************************************************************
** Color Primitives
******************************************************************************
*/

// Return the RGB value of a pixel in a color bitmap.

KV _CG(constcolmap c, int x, int y)
{
  byte *pb;

  if (gs.fErrorCheck) {
    if (!FValidate(c, x, y, "get"))
      return kvBlack;
  }
  pb = CJ(c, CI(c, x, y));
  return CK(pb);
}


// Set the RGB value of a pixel in a color bitmap.

void _CP(colmap c, int x, int y, KV kv)
{
  byte *pb;

  if (gs.fErrorCheck) {
    if (!FValidate(c, x, y, "put"))
      return;
    if (gs.fTraceDot && c == *gs.pbFocus)
      ScreenDot(x, y, fOff, kv);
  }
  pb = CJ(c, CI(c, x, y));
  CL(pb, RgbR(kv), RgbG(kv), RgbB(kv));
}


// Return the RGB value of a pixel in a color bitmap. Pixels not on the bitmap
// are assumed to be black.

KV Cg(constcolmap c, int x, int y)
{
  byte *pb;

  if (!FLegalC(c, x, y))
    return kvBlack;
  pb = CJ(c, CI(c, x, y));
  return CK(pb);
}


// Set the RGB value of a pixel in a color bitmap. Does nothing if an attempt
// is made to set a pixel not on the bitmap.

void Cp(colmap c, int x, int y, KV kv)
{
  byte *pb;

  if (!FLegalC(c, x, y))
    return;
  pb = CJ(c, CI(c, x, y));
  CL(pb, RgbR(kv), RgbG(kv), RgbB(kv));
}


// Allocate a new color bitmap of a given size.

flag FAllocateColmap(colmap *pc, int x, int y)
{
  colmap c;
  long cb;

  if (x < 0 || y < 0 || x > xColmap || y > yColmap) {
    PrintSzNN_E("Can't create color bitmap larger than %d by %d!\n",
      xColmap, yColmap);
    return fFalse;
  }
  cb = cbBH+CbColmap(x, y);
  if (cb < 0) {
    PrintSzNN_E("Can't allocate color bitmap of size %d by %d!\n", x, y);
    return fFalse;
  }
  *pc = (colmap)PAllocate(cb);
  c = *pc;
  if (c == NULL)
    return fFalse;
  CX(c) = x; CY(c) = y;
  c->clRow = CbColmapRow(x) >> 2;
  c->fColor = fTrue;
  c->wClient1 = wClient1Def; c->wClient2 = wClient2Def;
  c->wClient3 = wClient3Def; c->wClient4 = wClient4Def;
  return fTrue;
}


// Make a color bitmap be a given size, reallocating or allocating in the
// first place if necessary.

flag FColmapSizeSet(colmap *c, int xnew, int ynew)
{
  colmap c2;

  if (*c != NULL) {
    // Check if bitmap will have same dimensions, or same number of bytes.
    if (CX(*c) == xnew && CY(*c) == ynew)
      return fTrue;
    if (CbColmap(CX(*c), CY(*c)) == CbColmap(xnew, ynew)) {
      CX(*c) = xnew; CY(*c) = ynew;
      (*c)->clRow = CbColmapRow(xnew) >> 2;
      return fTrue;
    }
  }
  if (!FAllocateColmap(&c2, xnew, ynew))
    return fFalse;
  if (*c != NULL)
    DeallocateColmap(*c);
  *c = c2;
  return fTrue;
}


// Resize a color bitmap to be a given size, by either truncating or appending
// black pixels to its right and bottom edges.

flag FColmapResizeTo(colmap *c, int xnew, int ynew)
{
  colmap c2;

  if (CX(*c) == xnew && CY(*c) == ynew)
    return fTrue;
  if (!FAllocateColmap(&c2, xnew, ynew))
    return fFalse;
  ColmapSet(c2, kvBlack);
  if (!FNull(*c))
    BlockMoveC(*c, c2, 0, 0, CX(*c)-1, CY(*c)-1, 0, 0, -1);
  DeallocateColmap(*c);
  *c = c2;
  return fTrue;
}


// Resize a color bitmap by a given offset, by either truncating or appending
// black pixels to its left and top edges.

flag FColmapShiftBy(colmap *c, int xnew, int ynew)
{
  colmap c2;
  int x, y;

  if (xnew == 0 && ynew == 0)
    return fTrue;
  if (!FAllocateColmap(&c2, CX(*c) + xnew, CY(*c) + ynew))
    return fFalse;
  for (y = 0; y < CY(c2); y++)
    for (x = 0; x < CX(c2); x++)
      CP(c2, x, y, Cg(*c, x - xnew, y - ynew));
  DeallocateColmap(*c);
  *c = c2;
  return fTrue;
}


// Copy the contents of one color bitmap to another color bitmap.

flag FColmapCopy(constcolmap c1, colmap *c2)
{
  long clColmap, il;

  if (!FColmapSizeSet(c2, CX(c1), CY(c1)))
    return fFalse;
  CopyClient(c1, *c2);
  clColmap = CbColmap(CX(c1), CY(c1)) >> 2;
  for (il = 0; il < clColmap; il++)
    *MJ(*c2, il) = *MJ(c1, il);
  return fTrue;
}


// Lighten or darken an RGB color. The returned color ranges from black
// (r = -1.0) to the original color (r = 0.0) to white (r = 1.0).

KV KvShade(KV kv, real r)
{
  Assert(r >= -1.0 && r <= 1.0);
  if (r < 0.0) {
    return Rgb(
      (int)((real)(RgbR(kv))*(1.0 + r)),
      (int)((real)(RgbG(kv))*(1.0 + r)),
      (int)((real)(RgbB(kv))*(1.0 + r)));
  }
  return Rgb(
    255 - (int)((255.0 - (real)(RgbR(kv)))*(1.0 - r)),
    255 - (int)((255.0 - (real)(RgbG(kv)))*(1.0 - r)),
    255 - (int)((255.0 - (real)(RgbB(kv)))*(1.0 - r)));
}


// Blend two colors. Return a color half way between them.

KV KvBlend(KV kv1, KV kv2)
{
  return Rgb(
    (RgbR(kv1) + RgbR(kv2)) >> 1,
    (RgbG(kv1) + RgbG(kv2)) >> 1,
    (RgbB(kv1) + RgbB(kv2)) >> 1);
}


// Blend two colors with a specified bias given to one color. The returned
// color ranges from the first color (r = 0.0) to the second (r = 1.0).

KV KvBlendR(KV kv1, KV kv2, real r)
{
  Assert(FBetween(r, 0.0, 1.0));
  return Rgb((int)((real)(RgbR(kv2) - RgbR(kv1)) * r) + RgbR(kv1),
    (int)((real)(RgbG(kv2) - RgbG(kv1)) * r) + RgbG(kv1),
    (int)((real)(RgbB(kv2) - RgbB(kv1)) * r) + RgbB(kv1));
}


// Blend two colors with a specified bias given to one color. The returned
// color ranges from the first color (n1 = 0) to the second (n1 = n2).

KV KvBlendN(KV kv1, KV kv2, int n1, int n2)
{
  Assert(FBetween(n1, 0, n2));
  return Rgb((RgbR(kv2) - RgbR(kv1)) * n1 / n2 + RgbR(kv1),
    (RgbG(kv2) - RgbG(kv1)) * n1 / n2 + RgbG(kv1),
    (RgbB(kv2) - RgbB(kv1)) * n1 / n2 + RgbB(kv1));
}


// Return the difference between two RGB colors. Note the difference between a
// color and itself is black (like how X - X = 0), and the difference between
// a color and black is itself (like how X - 0 = X).

KV KvDiff(KV kv1, KV kv2)
{
  int nR, nG, nB;

  nR = NAbs(RgbR(kv2) - RgbR(kv1));
  nG = NAbs(RgbG(kv2) - RgbG(kv1));
  nB = NAbs(RgbB(kv2) - RgbB(kv1));
  return Rgb(nR, nG, nB);
}


// Return a color from a predefined range of colors.

KV Hue(int d)
{
  int nR, nG, nB, i;
  long l;

  l = (long)d * cs.nRainbowDistance / nDegMax +
    (long)cs.nRainbowStart * (nHueMax / nDegMax);

  // Do a blend between two colors.
  // d = 0 is color #1, d = 3600 is color #2
  if (!cs.fRainbowBlend) {
    while (l >= (nHueMax << 1))
      l -= (nHueMax << 1);
    while (l < 0)
      l += (nHueMax << 1);
    d = (int)l;
    if (d >= nHueMax)
      d = (nHueMax << 1) - 1 - d;
    return cs.fGraphNumber ? d :
      KvBlendN(cs.kvBlend1, cs.kvBlend2, d, nHueMax);
  }

  // Do a blend around the colors of the rainbow.
  // d = 0 is red, d = 1200 is green, d = 2400 is blue
  while (l >= nHueMax)
    l -= nHueMax;
  while (l < 0)
    l += nHueMax;
  d = (int)l;
  i = NAbs(d - nHueHalf);
  if (i > nHue13)
    nR = 255;
  else if (i < nHue16)
    nR = 0;
  else {
    i = i - nHue16;
    if (cs.fGraphNumber)
      i = (i + (Sq(nHue16) - Sq(nHue16 - i)) / nHue16) >> 1;
    nR = NMultDiv(i, 255, nHue16);
  }
  i = NAbs(d - nHue13);
  if (i < nHue16)
    nG = 255;
  else if (i > nHue13)
    nG = 0;
  else {
    i = nHue13 - i;
    if (cs.fGraphNumber)
      i = (i + (Sq(nHue16) - Sq(nHue16 - i)) / nHue16) >> 1;
    nG = NMultDiv(i, 255, nHue16);
  }
  i = NAbs(d - nHue23);
  if (i < nHue16)
    nB = 255;
  else if (i > nHue13)
    nB = 0;
  else {
    i = nHue13 - i;
    if (cs.fGraphNumber)
      i = (i + (Sq(nHue16) - Sq(nHue16 - i)) / nHue16) >> 1;
    nB = NMultDiv(i, 255, nHue16);
  }
  return Rgb(nR, nG, nB);
}


// Return a color given its Hue, Saturation, and Luminosity.

KV Hsl(int h, int s, int l)
{
  KV kv;

  kv = HueD(h);
  kv = KvShade(kv, (real)l / 1000.0);
  kv = KvBlendN(kvGray, kv, s, 1000);
  return kv;
}


/*
******************************************************************************
** Color Bitmap Editing
******************************************************************************
*/

// Set all pixels within a rectangle in a color bitmap to a color.

void BlockC(colmap c, int x1, int y1, int x2, int y2, KV kv)
{
  int y;

  Legalize(c, &x1, &y1); Legalize(c, &x2, &y2);
  SortN(&x1, &x2);
  SortN(&y1, &y2);
  for (y = y1; y <= y2; y++)
    LineXC(c, x1, x2, y, kv);
}


// Set pixels around the border of a rectangle in a color bitmap.

void BoxC(colmap c, int x1, int y1, int x2, int y2, int xsiz, int ysiz, KV kv)
{
  SortN(&x1, &x2);
  SortN(&y1, &y2);
  BlockC(c, x1, y1, x2, y1 + ysiz - 1, kv);
  if (y1 + ysiz < y2) {
    BlockC(c, x1, y1 + ysiz, x1 + xsiz - 1, y2 - ysiz, kv);
    BlockC(c, x2 - xsiz + 1, y1 + ysiz, x2, y2 - ysiz, kv);
  }
  BlockC(c, x1, y2 - ysiz + 1, x2, y2, kv);
}


// Copy pixels in a rectangle from one color bitmap to another. If the bitmaps
// are the same, the rectangles should not overlap.

void BlockMoveC(constcolmap c1, colmap c2, int x1, int y1, int x2, int y2,
  int x0, int y0, int nOp)
{
  int x, y, xT, yT;

  Legalize(c1, &x1, &y1); Legalize(c1, &x2, &y2);
  SortN(&x1, &x2);
  SortN(&y1, &y2);
  for (y = y1; y <= y2; y++)
    for (x = x1; x <= x2; x++) {
      xT = x0+(x-x1); yT = y0+(y-y1);
      if (FLegalC(c2, xT, yT))
        switch (nOp) {
        case 0: CP(c2, xT, yT, CG(c1, x, y) & CG(c2, xT, yT)); break;
        case 1: CP(c2, xT, yT, CG(c1, x, y) | CG(c2, xT, yT)); break;
        case 2: CP(c2, xT, yT, CG(c1, x, y) ^ CG(c2, xT, yT)); break;
        default: CP(c2, xT, yT, CG(c1, x, y)); break;
        }
    }
}


// Draw a horizontal line on a color bitmap.

void LineXC(colmap c, int x1, int x2, int y, KV kv)
{
  byte *pb;
  int x, nR, nG, nB;

  // Don't worry about pixels that are off the bitmap.
  if (y < 0 || y >= CY(c))
    return;
  SortN(&x1, &x2);
  if (x1 < 0)
    x1 = 0;
  if (x2 >= CX(c))
    x2 = CX(c)-1;

  // Quickly draw the line, where things like the memory address only need to
  // be calculated once.
  pb = CJ(c, CI(c, x1, y));
  nR = RgbR(kv); nG = RgbG(kv); nB = RgbB(kv);
  for (x = x1; x <= x2; x++) {
    CL(pb, nR, nG, nB);
    pb += cbPixelC;
  }
  if (gs.fTraceDot && c == *gs.pbFocus) {
    for (x = x1; x <= x2; x++)
      ScreenDot(x, y, fOff, kv);
  }
}


// Draw a vertical line on a color bitmap.

void LineYC(colmap c, int x, int y1, int y2, KV kv)
{
  byte *pb;
  int y, nR, nG, nB;

  // Don't worry about pixels that are off the bitmap.
  if (x < 0 || x >= CX(c))
    return;
  SortN(&y1, &y2);
  if (y1 < 0)
    y1 = 0;
  if (y2 >= CY(c))
    y2 = CY(c)-1;

  // Quickly draw the line, where things like the memory address only need to
  // be calculated once.
  pb = CJ(c, CI(c, x, y1));
  nR = RgbR(kv); nG = RgbG(kv); nB = RgbB(kv);
  for (y = y1; y <= y2; y++) {
    CL(pb, nR, nG, nB);
    pb += c->clRow << 2;
  }
  if (gs.fTraceDot && c == *gs.pbFocus) {
    for (y = y1; y <= y2; y++)
      ScreenDot(x, y, fOff, kv);
  }
}


// Draw a diagonal line of the specified tilt on a color bitmap.

void LineZC(colmap c, int x, int y1, int y2, int dir, KV kv)
{
  int i;

  for (i = 0; i <= y1 - y2; i++)
    CP(c, x + i*dir, y1 - i, kv);
}


// Draw a line between any two points on a color bitmap.

void LineC(colmap c, int x1, int y1, int x2, int y2, KV kv)
{
  int x = x1, y = y1, dx = x2 - x1, dy = y2 - y1, xInc, yInc, xInc2, yInc2,
    d, dInc, z, zMax;
  flag fAllLegal = FLegalC(c, x1, y1) && FLegalC(c, x2, y2);

  // Determine slope.
  if (NAbs(dx) >= NAbs(dy)) {
    xInc = NSgn(dx); yInc = 0;
    xInc2 = 0; yInc2 = NSgn(dy);
    zMax = NAbs(dx); dInc = NAbs(dy);
    d = zMax - (!FOdd(dx) && x1 > x2);
  } else {
    xInc = 0; yInc = NSgn(dy);
    xInc2 = NSgn(dx); yInc2 = 0;
    zMax = NAbs(dy); dInc = NAbs(dx);
    d = zMax - (!FOdd(dy) && y1 > y2);
  }
  d >>= 1;

  // Loop over long axis, adjusting short axis for slope as needed.
  for (z = 0; z <= zMax; z++) {
    if (fAllLegal || FLegalC(c, x, y))
      CP(c, x, y, kv);
    x += xInc; y += yInc; d += dInc;
    if (d >= zMax) {
      x += xInc2; y += yInc2; d -= zMax;
    }
  }
}


// Draw a filled in triangle on a color bitmap, defined by three points.

flag FTriangleC(colmap c, int x1, int y1, int x2, int y2, int x3, int y3,
  KV kv, int nTrans)
{
  MM rgmm[zBitmapStack], *pmm = NULL, *xminmax;
  int ymin, ymax, y, x, i;
  real rTrans = (real)nTrans / 10000.0;

  if (CY(c) <= zBitmapStack)
    xminmax = rgmm;
  else {
    pmm = RgAllocate(CY(c), MM);
    if (pmm == NULL)
      return fFalse;
    xminmax = pmm;
  }
  ymin = NMax(NMin(Min(y1, y2), y3), 0);
  ymax = NMin(NMax(Max(y1, y2), y3), CY(c) - 1);
  for (y = ymin; y <= ymax; y++) {
    xminmax[y].min = xColmap;
    xminmax[y].max = -1;
  }
  LineSetup(c, xminmax, x1, y1, x2, y2);
  LineSetup(c, xminmax, x2, y2, x3, y3);
  LineSetup(c, xminmax, x3, y3, x1, y1);

  // Draw one horizontal line per row.
  if (rTrans <= 0.0) {
    for (y = ymin; y <= ymax; y++)
      LineXC(c, xminmax[y].min, xminmax[y].max, y, kv);
  } else {
    for (y = ymin; y <= ymax; y++) {
      i = Min(xminmax[y].max, CX(c) - 1);
      for (x = Max(xminmax[y].min, 0); x <= i; x++)
        CP(c, x, y, KvBlendR(kv, CG(c, x, y), rTrans));
    }
  }
  if (pmm != NULL)
    DeallocateP(pmm);
  return fTrue;
}


// Draw a filled in four sided quadrilateral on a color bitmap.

flag FQuadrilateralC(colmap c, int x1, int y1, int x2, int y2, int x3, int y3,
  int x4, int y4, KV kv, int nTrans)
{
  MM rgmm[zBitmapStack], *pmm = NULL, *xminmax;
  int ymin, ymax, y, x, i;
  real rTrans = (real)nTrans / 10000.0;

  if (CY(c) <= zBitmapStack)
    xminmax = rgmm;
  else {
    pmm = RgAllocate(CY(c), MM);
    if (pmm == NULL)
      return fFalse;
    xminmax = pmm;
  }
  ymin = NMax(NMin(NMin(Min(y1, y2), y3), y4), 0);
  ymax = NMin(NMax(NMax(Max(y1, y2), y3), y4), CY(c) - 1);
  for (y = ymin; y <= ymax; y++) {
    xminmax[y].min = xColmap;
    xminmax[y].max = -1;
  }
  LineSetup(c, xminmax, x1, y1, x2, y2);
  LineSetup(c, xminmax, x2, y2, x3, y3);
  LineSetup(c, xminmax, x3, y3, x4, y4);
  LineSetup(c, xminmax, x4, y4, x1, y1);

  // Since one horizontal line is drawn per row, the quadrilateral should be
  // convex, or at least not concave any along the vertical axis.
  if (rTrans <= 0.0) {
    for (y = ymin; y <= ymax; y++)
      LineXC(c, xminmax[y].min, xminmax[y].max, y, kv);
  } else {
    for (y = ymin; y <= ymax; y++) {
      i = Min(xminmax[y].max, CX(c) - 1);
      for (x = Max(xminmax[y].min, 0); x <= i; x++)
        CP(c, x, y, KvBlendR(kv, CG(c, x, y), rTrans));
    }
  }
  if (pmm != NULL)
    DeallocateP(pmm);
  return fTrue;
}


// Draw a filled in circle or ellipse, within a rectangle on a color bitmap.

void DiskC(colmap c, int x1, int y1, int x2, int y2, KV kv)
{
  int rx, ry, r, i, j, iT, jT, q, qLo, qHi, m1, m2, n1, n2;

  rx = NAbs(x2 - x1) >> 1; ry = NAbs(y2 - y1) >> 1;
  if (rx == 0 || ry == 0) {
    BlockC(c, x1, y1, x2, y2, kv);
    return;
  }
  r = Max(rx, ry);
  j = r; q = r+1; qLo = 1; qHi = r+r-1;
  m1 = (x1 + x2) >> 1; m2 = m1 + FOdd(x2 - x1);
  n1 = (y1 + y2) >> 1; n2 = n1 + FOdd(y2 - y1);
  for (i = 0; i <= r; i++) {
    iT = (rx <= ry) ? i : i * ry / rx;
    jT = (rx >= ry) ? j : j * rx / ry;
    LineXC(c, m1 - jT, m2 + jT, n1 - iT, kv);
    LineXC(c, m1 - jT, m2 + jT, n2 + iT, kv);
    q -= qLo;
    while (q < 0) {
      q += qHi;
      qHi -= 2;
      j--;
    }
    qLo += 2;
  }
}


// Draw an outline of a circle or ellipse, within a rectangle on a bitmap.

void CircleC(colmap c, int x1, int y1, int x2, int y2, KV kv)
{
  int rx, ry, r, imax, i, j, i1, i2, j1, j2, q, qLo, qHi, m1, m2, n1, n2;

  // This efficient algorithm has no calculations more complicated than
  // addition (and multiplication to proportion in the case of an ellipse).
  rx = NAbs(x2 - x1) >> 1; ry = NAbs(y2 - y1) >> 1;
  if (rx == 0 || ry == 0) {
    BlockC(c, x1, y1, x2, y2, kv);
    return;
  }
  r = Max(rx, ry);
  j = r; q = r+1; qLo = 1; qHi = r+r-1;
  m1 = (x1 + x2) >> 1; m2 = m1 + FOdd(x2 - x1);
  n1 = (y1 + y2) >> 1; n2 = n1 + FOdd(y2 - y1);
  imax = NMultDiv(r + 1, 708, 1000);
  for (i = 0; i <= imax; i++) {
    i1 = j2 = j; j1 = i2 = i;
    if (rx < ry) {
      i1 = i1 * rx / ry; i2 = i2 * rx / ry;
    } else if (ry < rx) {
      j1 = j1 * ry / rx; j2 = j2 * ry / rx;
    }
    Cp(c, m1 - i1, n1 - j1, kv); Cp(c, m1 - i2, n1 - j2, kv);
    Cp(c, m2 + i1, n1 - j1, kv); Cp(c, m2 + i2, n1 - j2, kv);
    Cp(c, m1 - i1, n2 + j1, kv); Cp(c, m1 - i2, n2 + j2, kv);
    Cp(c, m2 + i1, n2 + j1, kv); Cp(c, m2 + i2, n2 + j2, kv);
    q -= qLo;
    if (q < 0) {
      q += qHi;
      qHi -= 2;
      j--;
    }
    qLo += 2;
  }
}


// Draw an arc of an ellipse on a color bitmap, specifically some subsection
// of sides of a dMax sized polygon.

void ArcPolygonC(colmap c, int x, int y, int rx, int ry,
  int d1, int d2, int dInc, int dMax, KV kv)
{
  int i, m, n, m0, n0;
  real rMax;

  if (dMax < 0) {
    dMax = 64;
    if (rx + ry > 384)
      dMax = 256;
    d1 = 0;
    d2 = dMax;
  }
  rMax = (real)dMax;
  for (i = d1; i <= d2; i += dInc) {
    AngleR(&m, &n, x, y, (real)rx, (real)ry, (real)i * rDegMax / rMax);
    if (i > d1)
      LineC(c, m0, n0, m, n, kv);
    m0 = m; n0 = n;
  }
}


#define FLegalFillC(c, x, y, kv) (FLegalC(c, x, y) && CG(c, x, y) == (kv))

// Flood an irregular shape in a color bitmap with one color, where the shape
// area is defined by another color.

flag FFillC(colmap c, int x, int y, KV kv, KV kvArea, flag fFlood)
{
  PT *rgpt;
  long ipt = 0;
  int xnew, ynew, d, dMax;

  if (!FLegalC(c, x, y))
    return fTrue;
  if (F64K(c))
    return fFalse;
  if (kvArea < 1)
    kvArea = CG(c, x, y);
  if (kv == kvArea || CG(c, x, y) != kvArea)
    return fTrue;
  rgpt = RgAllocate(CX(c)*CY(c), PT);
  if (rgpt == NULL)
    return fFalse;
  dMax = DIRS + fFlood*DIRS;
LSet:
  CP(c, x, y, kv);
LNext:
  for (d = 0; d < dMax; d++) {
    xnew = x + xoff[d]; ynew = y + yoff[d];
    if (FLegalFillC(c, xnew, ynew, kvArea)) {
      FillPush(x, y);
      x = xnew; y = ynew;
      goto LSet;
    }
  }
  if (ipt > 0) {
    FillPop(x, y);
    goto LNext;
  }
  DeallocateP(rgpt);
  return fTrue;
}


/*
******************************************************************************
** Color Bitmap Transformations
******************************************************************************
*/

// Set all pixels in a color bitmap to a specified color.

void ColmapSet(colmap c, KV kv)
{
  int il;
  dword *pl, l;
  flag fTrace;

  if (kv == kvBlack || kv == kvWhite) {

    // For the simplest colors, set 32 bits at a time.
    l = kv == kvBlack ? 0 : ~0;
    pl = (dword *)CJ(c, 0);
    for (il = CY(c) * c->clRow; il > 0; il--)
      *pl++ = l;
  } else {

    // For normal colors, go set one pixel at a time.
    fTrace = gs.fTraceDot; gs.fTraceDot = fFalse;
    BlockC(c, 0, 0, CX(c)-1, CY(c)-1, kv);
    gs.fTraceDot = fTrace;
  }
  if (gs.fTraceDot && c == *gs.pbFocus)
    UpdateDisplay();
}


// Convert all colors in a color bitmap to shades of gray with the same
// brightness.

void ColmapGrayscale(colmap c)
{
  int x, y, n, nR, nG, nB, nRT, nGT, nBT;
  byte *pb;

  // Simple case: Treat RGB channels as the same brightness.
  if (cs.lGrayscale < 0) {
    for (y = 0; y < CY(c); y++) {
      pb = CJ(c, CI(c, 0, y));
      for (x = 0; x < CX(c); x++) {
        CM(pb, nR, nG, nB);
        n = (nR + nG + nB + 1) / 3;
        CL(pb, n, n, n);
        pb += cbPixelC;
      }
    }

  // Advanced case: Each of the RGB channels has its own brightness.
  } else {
    nRT = cs.lGrayscale / 1000000;
    nGT = cs.lGrayscale / 1000 % 1000;
    nBT = cs.lGrayscale % 1000;
    for (y = 0; y < CY(c); y++) {
      pb = CJ(c, CI(c, 0, y));
      for (x = 0; x < CX(c); x++) {
        CM(pb, nR, nG, nB);
        nR = nR * nRT / 1000;
        nG = nG * nGT / 1000;
        nB = nB * nBT / 1000;
        n = nR + nG + nB;
        CL(pb, n, n, n);
        pb += cbPixelC;
      }
    }
  }
}


// Adjust the brightness of the colors in a color bitmap.

void ColmapBrightness(colmap c, int nOp, real rFactor, int nLimit)
{
  int x, y, r, g, b, sum;
  KV kv;

  switch (nOp) {

  // Brighten all by factor.
  case 0:
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++) {
        CP(c, x, y, KvShade(CG(c, x, y), rFactor));
      }
    break;

  // Brighten all by offset.
  case 1:
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++) {
        kv = CG(c, x, y);
        r = RgbR(kv); g = RgbG(kv); b = RgbB(kv);
        if (nLimit >= 0) {
          r = Min(r + nLimit, 255);
          g = Min(g + nLimit, 255);
          b = Min(b + nLimit, 255);
        } else {
          r = Max(r + nLimit, 0);
          g = Max(g + nLimit, 0);
          b = Max(b + nLimit, 0);
        }
        CP(c, x, y, Rgb(r, g, b));
      }
    break;

  // Brighten all by multiplier.
  case 2:
    if (!cs.fGraphNumber) {
      for (y = 0; y < CY(c); y++)
        for (x = 0; x < CX(c); x++) {
          kv = CG(c, x, y);
          r = RgbR(kv); g = RgbG(kv); b = RgbB(kv);
          if (rFactor >= 0.0) {
            r = (int)Min((long)((real)r * rFactor), 255);
            g = (int)Min((long)((real)g * rFactor), 255);
            b = (int)Min((long)((real)b * rFactor), 255);
          } else {
            r = 255-(int)Min((long)((real)(255-r) * -rFactor), 255);
            g = 255-(int)Min((long)((real)(255-g) * -rFactor), 255);
            b = 255-(int)Min((long)((real)(255-b) * -rFactor), 255);
          }
          CP(c, x, y, Rgb(r, g, b));
        }
    } else {
      for (y = 0; y < CY(c); y++)
        for (x = 0; x < CX(c); x++) {
          kv = CG(c, x, y);
          kv = (KV)((real)kv * rFactor);
          CP(c, x, y, kv);
        }
    }
    break;

  // Limit brightness to value.
  case 3:
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++) {
        kv = CG(c, x, y);
        r = RgbR(kv); g = RgbG(kv); b = RgbB(kv);
        sum = r + g + b;
        if (nLimit >= 0) {
          if (sum > nLimit) {
            r = NMultDiv(r, nLimit, sum);
            g = NMultDiv(g, nLimit, sum);
            b = NMultDiv(b, nLimit, sum);
          }
        } else {
          if (sum < -nLimit) {
            r = 255-NMultDiv(255-r, 255*3 + nLimit, 255*3 - sum);
            g = 255-NMultDiv(255-g, 255*3 + nLimit, 255*3 - sum);
            b = 255-NMultDiv(255-b, 255*3 + nLimit, 255*3 - sum);
          }
        }
        CP(c, x, y, Rgb(r, g, b));
      }
    break;

  default:
    Assert(fFalse);
  }
}


// Replace one color in a bitmap with another color or a pattern.

void ColmapReplace(colmap c, KV kvFrom, KV kvTo,
  int x1, int y1, int x2, int y2)
{
  int x, y;
  byte *pb, nR, nG, nB;

  if (kvTo < 0 && kvTo > -crpUnknown) {
    ColmapReplacePattern(c, kvFrom, (int)(-kvTo), x1, y1, x2, y2);
    return;
  }

  // Quickly replace all instances of one color with another.
  nR = RgbR(kvTo); nG = RgbG(kvTo); nB = RgbB(kvTo);
  for (y = 0; y < CY(c); y++) {
    pb = CJ(c, CI(c, 0, y));
    for (x = 0; x < CX(c); x++) {
      if (CK(pb) == kvFrom) {
        CL(pb, nR, nG, nB);
      }
      pb += cbPixelC;
    }
  }
}


// Replace one color in a bitmap with a special color pattern.

void ColmapReplacePattern(colmap c, KV kvFrom, int nMode,
  int x1, int y1, int x2, int y2)
{
  int x, y, h, v, l;
  KV kv;
  real len;

  h = CX(c) >> 1;
  v = CY(c) >> 1;
  switch (nMode) {
  case crpRing: // O
    len = LengthN(CX(c), CY(c), h, v) / rHueMax;
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, Hue((int)(LengthN(h, v, x, y) / len + rRound)));
    break;
  case crpSpoke: // X
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, HueR(GetAngle(h, v, x, y)));
    break;
  case crpFade: // *
    len = LengthN(CX(c), CY(c), h, v) + rMinute;
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, KvShade(HueR(GetAngle(h, v, x, y)),
            (0.5 - LengthN(x, y, h, v) / len) * 2.0));
    break;
  case crpY: // |
    for (y = 0; y < CY(c); y++) {
      kv = Hue(NMultDiv(y, nHueMax, CY(c)));
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, kv);
    }
    break;
  case crpX: // -
    for (x = 0; x < CX(c); x++) {
      kv = Hue(NMultDiv(x, nHueMax, CX(c)));
      for (y = 0; y < CY(c); y++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, kv);
    }
    break;
  case crpDiagUL: // \ 
    h = CX(c) + CY(c);
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, Hue(NMultDiv(x + y, nHueMax, h)));
    break;
  case crpDiagUR: // /
    h = CX(c) + CY(c);
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, Hue(NMultDiv(x + CY(c)-y-1, nHueMax, h)));
    break;
  case crpSquare: // #
    l = Max(h, v);
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, Hue(NMultDiv(Max(NAbs(h-x), NAbs(v-y)), nHueMax, l)));
    break;
  case crpDiamond: // +
    l = Max(h, v) << 1;
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, Hue(NMultDiv(NAbs(h-x) + NAbs(v-y), nHueMax, l)));
    break;
  case crpSpiral: // @
    len = LengthN(x1, y1, h, v);
    if (len <= 0)
      len = rDegMax;
    if (y1 > v || (y1 == v && x1 > h))
      neg(len);
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, HueR(GetAngle(h, v, x, y) +
            LengthN(h, v, x, y) / len * rDegMax));
    break;
  case crpRing2: // o
    len = LengthN(x1, y1, 0, 0);
    len = RMax(len, LengthN(x1, y1, CX(c)-1, 0));
    len = RMax(len, LengthN(x1, y1, 0, CY(c)-1));
    len = RMax(len, LengthN(x1, y1, CX(c)-1, CY(c)-1)) / rHueMax;
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, Hue((int)(LengthN(x1, y1, x, y) / len + rRound)));
    break;
  case crpRings: // 8
    len = RMin(LengthN(x1, y1, 0, 0), LengthN(x2, y2, 0, 0));
    len = RMax(len, RMin(LengthN(x1, y1, CX(c)-1, 0),
      LengthN(x2, y2, CX(c)-1, 0)));
    len = RMax(len, RMin(LengthN(x1, y1, 0, CY(c)-1),
      LengthN(x2, y2, 0, CY(c)-1)));
    len = RMax(len, RMin(LengthN(x1, y1, CX(c)-1, CY(c)-1),
      LengthN(x2, y2, CX(c)-1, CY(c)-1))) / rHueMax;
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, Hue((int)(RMin(LengthN(x1, y1, x, y),
            LengthN(x2, y2, x, y)) / len + rRound)));
    break;
  case crpSpoke2: // x
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, HueR(GetAngle(x1, y1, x, y)));
    break;
  case crpRainbow: // ?
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, kvRandomRainbow);
    break;
  case crpRandom: // ??
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, kvRandom);
    break;
  case crpRandom2: // ?????
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) == kvFrom)
          CP(c, x, y, Rgb(Rnd(RgbR(cs.kvBlend1), RgbR(cs.kvBlend2)),
            Rnd(RgbG(cs.kvBlend1), RgbG(cs.kvBlend2)),
            Rnd(RgbB(cs.kvBlend1), RgbB(cs.kvBlend2))));
    break;
  default:
    Assert(fFalse);
  }
}


// Exchange two colors in a color bitmap with each other.

void ColmapExchange(colmap c, KV kv1, KV kv2)
{
  int x, y;
  KV kv;

  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++) {
      kv = CG(c, x, y);
      if (kv == kv1)
        CP(c, x, y, kv2);
      else if (kv == kv2)
        CP(c, x, y, kv1);
    }
}


// Combine a color with all pixels in a bitmap using the specified operator.

void ColmapOrAndKv(colmap c, KV kv, int nOp)
{
  int x, y, nR, nG, nB;
  KV kvT;

  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++) {
      kvT = CG(c, x, y);
      switch (nOp) {

      // Boolean bit combining
      case 0: kvT &= kv; break;
      case 1: kvT |= kv; break;
      case 2: kvT ^= kv; break;

      // Add/Sub/Average (separate rgb channels)
      case 3: // +
        nR = NMin(RgbR(kvT) + RgbR(kv), 255);
        nG = NMin(RgbG(kvT) + RgbG(kv), 255);
        nB = NMin(RgbB(kvT) + RgbB(kv), 255);
        kvT = Rgb(nR, nG, nB);
        break;
      case 4: kvT = KvDiff(kvT, kv);  break;
      case 5: kvT = KvBlend(kvT, kv); break;

      // Add/Sub/Average (single number)
      case 6: kvT = NMin(kvT + kv, kvWhite); break;
      case 7: kvT = NMax(kvT - kv, kvBlack); break;
      case 8: kvT = (kvT + kv) >> 1; break;
      case 9:  kvT = NMin(kvT, kv); break;
      case 10: kvT = NMax(kvT, kv); break;
      default: Assert(fFalse);
      }
      CP(c, x, y, kvT);
    }
}


// Combine one color bitmap with another using the specified operator.

void ColmapOrAnd(colmap c1, constcolmap c2, int nOp)
{
  int x, y, nR, nG, nB;
  KV kv1, kv2, kv;

  for (y = 0; y < CY(c1); y++)
    for (x = 0; x < CX(c1); x++) {
      kv1 = CG(c1, x, y); kv2 = Cg(c2, x, y);
      switch (nOp) {
      case 0: kv = kv1 & kv2; break;
      case 1: kv = kv1 | kv2; break;
      case 2: kv = kv1 ^ kv2; break;
      case 3: // +
        nR = NMin(RgbR(kv1) + RgbR(kv2), 255);
        nG = NMin(RgbG(kv1) + RgbG(kv2), 255);
        nB = NMin(RgbB(kv1) + RgbB(kv2), 255);
        kv = Rgb(nR, nG, nB);
        break;
      case 4: kv = KvDiff(kv1, kv2);  break;
      case 5: kv = KvBlend(kv1, kv2); break;
      case 6: kv = NMin(kv1 + kv2, kvWhite); break;
      case 7: kv = NMax(kv1 - kv2, kvBlack); break;
      case 8: kv = (kv1 + kv2) >> 1; break;
      default: Assert(fFalse);
      }
      CP(c1, x, y, kv);
    }
}


// Blend one color bitmap with another, based on the transparency proportions
// in a third bitmap.

void ColmapAlpha(colmap c1, constcolmap c2, constcolmap c3)
{
  int x, y, nR, nG, nB;
  KV kv1, kv2, kv3;

  for (y = 0; y < CY(c1); y++)
    for (x = 0; x < CX(c1); x++) {
      kv1 = CG(c1, x, y); kv2 = Cg(c2, x, y); kv3 = Cg(c3, x, y);
      nR = (RgbR(kv2) - RgbR(kv1)) * RgbR(kv3) / 255 + RgbR(kv1);
      nG = (RgbG(kv2) - RgbG(kv1)) * RgbG(kv3) / 255 + RgbG(kv1);
      nB = (RgbB(kv2) - RgbB(kv1)) * RgbB(kv3) / 255 + RgbB(kv1);
      CP(c1, x, y, Rgb(nR, nG, nB));
    }
}


// Tile one color bitmap across and down within the size of another bitmap.

void ColmapTessellate(colmap c1, constcolmap c2)
{
  int x, y;

  for (y = 0; y < CY(c1); y += CY(c2))
    for (x = 0; x < CX(c1); x += CX(c2))
      BlockMoveC(c2, c1, 0, 0, CX(c2)-1, CY(c2)-1, x, y, -1);
}


// Resize a color bitmap to the specified size, then tessellate the old
// pattern within the bounds of the new dimensions.

flag FColmapTessellateTo(colmap *c, int x, int y)
{
  colmap c2;

  if (!FAllocateColmap(&c2, x, y))
    return fFalse;
  ColmapTessellate(c2, *c);
  DeallocateColmap(*c);
  *c = c2;
  return fTrue;
}


// Flip the pixels in a color bitmap horizontally.

void ColmapFlipX(colmap c)
{
  int x, y;
  KV kv;

  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c) >> 1; x++) {
      kv = CG(c, x, y);
      CP(c, x, y, CG(c, CX(c)-1-x, y));
      CP(c, CX(c)-1-x, y, kv);
    }
}


// Flip the pixels in a color bitmap vertically.

void ColmapFlipY(colmap c)
{
  int x, y;
  KV kv;

  for (x = 0; x < CX(c); x++)
    for (y = 0; y < CY(c) >> 1; y++) {
      kv = CG(c, x, y);
      CP(c, x, y, CG(c, x, CY(c)-1-y));
      CP(c, x, CY(c)-1-y, kv);
    }
}


// Transpose the pixels in a color bitmap i.e. flip across the main diagonal.

flag FColmapTranspose(colmap *pc)
{
  colmap c = *pc, c2;
  int x, y;
  KV kv;

  if (CX(c) == CY(c)) {

    // For square bitmaps, just swap the pixels across the main diagonal.
    for (y = 1; y < CY(c); y++)
      for (x = 0; x < y; x++) {
        kv = CG(c, x, y);
        CP(c, x, y, CG(c, y, x));
        CP(c, y, x, kv);
      }
  } else {

    // For rectangular bitmaps, create a new bitmap of the transposed size.
    if (!FAllocateColmap(&c2, CY(c), CX(c)))
      return fFalse;
    for (y = 0; y < CY(c2); y++)
      for (x = 0; x < CX(c2); x++)
        CP(c2, x, y, CG(c, y, x));
    DeallocateColmap(c);
    *pc = c2;
  }
  return fTrue;
}


// Resize a color bitmap to be a given size. The old contents of the bitmap
// are stretched or compacted to fit within the new dimensions.

flag FColmapZoomTo(colmap *c, int xnew, int ynew, flag fPreserve)
{
  colmap c2;
  int x, y, x1, y1, x2, y2, nR, nG, nB, i;
  real deltax, deltay;
  KV kv;

  if (!FAllocateColmap(&c2, xnew, ynew))
    return fFalse;
  if (FNull(*c) || FNull(c2))
    goto LDone;
  deltax = (real)CX(*c) / (real)xnew;
  deltay = (real)CY(*c) / (real)ynew;
  if (fPreserve && deltax >= 1.0 && deltay >= 1.0) {

    // Preserve on pixels when shrinking case: If the bitmap isn't growing in
    // either dimension, then there are always one or more source pixels to
    // map to a destination pixel. Make each destination pixel the mixture of
    // colors of all the source pixels that map to it.

    for (y = 0; y < ynew; y++)
      for (x = 0; x < xnew; x++) {
        nR = nG = nB = i = 0;
        y2 = (int)((real)(y+1)*deltay);
        for (y1 = (int)((real)y*deltay); y1 < y2; y1++) {
          x2 = (int)((real)(x+1)*deltax);
          for (x1 = (int)((real)x*deltax); x1 < x2; x1++) {
            kv = CG(*c, x1, y1);
            nR += RgbR(kv); nG += RgbG(kv); nB += RgbB(kv);
            i++;
          }
        }
        CP(c2, x, y,
          Rgb((nR + (i >> 1)) / i, (nG + (i >> 1)) / i, (nB + (i >> 1)) / i));
      }
  } else {

    // Normal case: Make each destination pixel a copy of the one source pixel
    // that best maps to it.

    for (y = 0; y < ynew; y++)
      for (x = 0; x < xnew; x++)
        CP(c2, x, y, CG(*c, (int)((real)x*deltax + rMinute),
          (int)((real)y*deltay + rMinute)));
  }
LDone:
  DeallocateColmap(*c);
  *c = c2;
  return fTrue;
}


// Zoom a color bitmap and its contents by the given factors. Each source
// pixel becomes an x by y block in the destination, where x and y can be
// different for even and odd numbered rows and columns.

flag FColmapBias(colmap *pc, int xbias1, int xbias2, int ybias1, int ybias2)
{
  colmap c = *pc, c2;
  int x, y;

  if (!FAllocateColmap(&c2,
    (CX(c) >> 1)*(xbias1 + xbias2) + FOdd(CX(c))*xbias1,
    (CY(c) >> 1)*(ybias1 + ybias2) + FOdd(CY(c))*ybias1))
    return fFalse;
  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++)
      BlockC(c2, (x >> 1)*(xbias1 + xbias2) + FOdd(x)*xbias1,
        (y >> 1)*(ybias1 + ybias2) + FOdd(y)*ybias1,
        ((x + 1) >> 1)*(xbias1 + xbias2) + FOdd(x + 1)*xbias1 - 1,
        ((y + 1) >> 1)*(ybias1 + ybias2) + FOdd(y + 1)*ybias1 - 1,
        CG(c, x, y));
  DeallocateBitmap(c);
  *pc = c2;
  return fTrue;
}


// Resize a color bitmap smaller, by trimming all black color rows and columns
// from its four edges.

flag FColmapCollapse(colmap *c)
{
  int xnew, ynew, i;

  // Trim black pixels from the right and bottom edges.
  for (ynew = CY(*c)-1; ynew >= 0; ynew--)
    for (i = 0; i < CX(*c); i++)
      if (CG(*c, i, ynew) != kvBlack)
        goto LNext1;
LNext1:
  for (xnew = CX(*c)-1; xnew >= 0; xnew--)
    for (i = 0; i <= ynew; i++)
      if (CG(*c, xnew, i) != kvBlack)
        goto LNext2;
LNext2:
  if (!FColmapResizeTo(c, xnew + 1, ynew + 1))
    return fFalse;

  // Trim black pixels from the left and top edges.
  for (ynew = 0; ynew < CY(*c); ynew++)
    for (i = 0; i < CX(*c); i++)
      if (CG(*c, i, ynew) != kvBlack)
        goto LNext3;
LNext3:
  for (xnew = 0; xnew < CX(*c); xnew++)
    for (i = ynew; i < CY(*c); i++)
      if (CG(*c, xnew, i) != kvBlack)
        goto LNext4;
LNext4:
  return FColmapShiftBy(c, -xnew, -ynew);
}


// Blur the colors in a bitmap together slightly.

flag FColmapBlur(colmap c)
{
  colmap c2 = NULL;
  int x, y, xnew, ynew, d, cd, nR, nG, nB, nRT, nGT, nBT;
  byte *pb, *pb2, *pbT;

  if (!FColmapCopy(c, &c2))
    return fFalse;
  for (y = 0; y < CY(c2); y++) {
    pb = CJ(c, CI(c, 0, y)); pb2 = CJ(c2, CI(c2, 0, y));
    for (x = 0; x < CX(c2); x++) {

      // For each pixel, average its color with its neighboring pixels.
      cd = 1;
      CM(pb2, nR, nG, nB);
      for (d = 0; d < DIRS2; d++) {
        xnew = x + xoff[d]; ynew = y + yoff[d];
        if (FLegalC(c2, xnew, ynew)) {
          cd++;
          pbT = CJ(c2, CI(c2, xnew, ynew));
          CM(pbT, nRT, nGT, nBT);
          nR += nRT; nG += nGT; nB += nBT;
        }
      }
      d = cd >> 1;
      nR = (nR + d) / cd; nG = (nG + d) / cd; nB = (nB + d) / cd;
      CL(pb, nR, nG, nB);
      pb += cbPixelC; pb2 += cbPixelC;
    }
  }
  DeallocateColmap(c2);
  return fTrue;
}


// Make a color bitmap have as high contrast as possible.

void ColmapContrast(colmap c, flag fRainbow)
{
  int x, y, nR, nG, nB, rgn[3], i, iMin, iMax, nMin, nMax;
  byte *pb;

  if (!fRainbow) {

    // Standard contrast: Each pixel maps to the nearest of one of eight
    // colors, where all RGB values become 0 or 255.

    for (y = 0; y < CY(c); y++) {
      pb = CJ(c, CI(c, 0, y));
      for (x = 0; x < CX(c); x++) {
        CM(pb, nR, nG, nB);
        nR = nR < 128 ? 0 : 255;
        nG = nG < 128 ? 0 : 255;
        nB = nB < 128 ? 0 : 255;
        CL(pb, nR, nG, nB);
        pb += cbPixelC;
      }
    }
  } else {

    // Rainbow contrast: Each pixel maps to the nearest rainbow color to it.
    // A rainbow color has at least one RGB component of 255 and another 0,
    // so increase the highest RGB component to 255 and drop the lowest to 0.

    for (y = 0; y < CY(c); y++) {
      pb = CJ(c, CI(c, 0, y));
      for (x = 0; x < CX(c); x++) {
        CM(pb, rgn[0], rgn[1], rgn[2]);
        nMax = -1; nMin = 256;
        for (i = 0; i < 3; i++)
          if (rgn[i] > nMax) {
            nMax = rgn[i]; iMax = i;
          }
        for (i = 0; i < 3; i++)
          if (rgn[i] <= nMin) {
            nMin = rgn[i]; iMin = i;
          }
        rgn[iMax] = 255; rgn[iMin] = 0;
        CL(pb, rgn[0], rgn[1], rgn[2]);
        pb += cbPixelC;
      }
    }
  }
}


// Transform the pixels in a color bitmap in some special manner.

flag FColmapTransform(colmap c, int nOp, int nFactor)
{
  colmap c2 = NULL;
  int xmid, ymid, x, y, x2, y2;
  real rFactor = (real)-nFactor, rx, ry, rS, rC, rl, rlMax;

  if (!FColmapCopy(c, &c2))
    return fFalse;
  xmid = CX(c) >> 1; ymid = CY(c) >> 1;
  rlMax = (real)Min(xmid, ymid);
  switch (nOp) {

  // Rotate all by degree.
  case 0:
    RotateR2Init(rS, rC, rFactor);
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++) {
        rx = (real)(x - xmid); ry = (real)(y - ymid);
        RotateR2(&rx, &ry, rS, rC);
        x2 = xmid + (int)rx; y2 = ymid + (int)ry;
        CP(c, x, y, Cg(c2, x2, y2));
      }
    break;

  // Twist middle by degree.
  case 1:
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++) {
        rx = (real)(x - xmid); ry = (real)(y - ymid);
        rl = LengthR(0, 0, rx, ry);
        if (rl >= rlMax)
          continue;
        rl = 1.0 - (rl / rlMax);
        RotateR(&rx, &ry, Sq(rl) * rFactor);
        x2 = xmid + (int)rx; y2 = ymid + (int)ry;
        CP(c, x, y, Cg(c2, x2, y2));
      }
    break;

  // Replace number with hue.
  case 2:
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        CP(c, x, y, Hue(CG(c2, x, y) * nFactor / 100 % nHueMax));
    break;

  default:
    Assert(fFalse);
  }
  DeallocateColmap(c2);
  return fTrue;
}


CONST int rgnSprite[16] =
  {15, 0, 1, 4, 2, 12, 5, 9, 3, 7, 13, 8, 6, 11, 10, 14};

// Zoom a bitmap by a certain magnification based on the size and using the
// contents of a second bitmap. Use the status of a pixel and its four
// neighbors in a monochrome bitmap to determine which section of the second
// bitmap to use for the zoomed version of each pixel.

flag FZoomSprite(bitmap *b1, colmap *c1, constbitmap b2, constcolmap c2)
{
  bitmap b, bT = NULL;
  colmap c;
  flag fColor = (c2 != NULL);
  int x1 = MX(*b1), y1 = MY(*b1), x2, y2, xd, yd, x, y, d, n, zp;

  x2 = !fColor ? MX(b2) : CX(c2);
  y2 = !fColor ? MY(b2) : CY(c2);
  zp = x2 >> 2;
  if (x2 < 1 || y2 < 1 || (x2 & 3) != 0 || y2 % zp != 0)
    return fFalse;
  if (!fColor) {
    if (!FBitmapCopy(*b1, &bT))
      return fFalse;
    if (!FBitmapSizeSet(b1, x1 * zp, y1 * zp)) {
      DeallocateBitmap(bT);
      return fFalse;
    }
    b = *b1;
    BitmapOff(b);
  } else {
    bT = *b1;
    if (!FColmapSizeSet(c1, x1 * zp, y1 * zp))
      return fFalse;
    c = *c1;
    ColmapSet(*c1, kvBlack);
  }
  for (y = 0; y < y1; y++)
    for (x = 0; x < x1; x++) {
      for (n = 0, d = DIRS1; d >= 0; d--)
        n = (n << 1) | Mg(bT, x + xoff[d], y + yoff[d]);
      n = rgnSprite[n] + (!Mg(bT, x, y) * 16);
      xd = (n & 3) * zp; yd = (n >> 2) * zp;

      // Copy the section to the destination. If the second bitmap doesn't
      // have enough rows, use a solid black or white block.

      if (!fColor) {
        if (yd >= y2) {
          if (yd < x2)
            Block(b, x * zp, y * zp, x * zp + zp-1, y * zp + zp-1, fOn);
          continue;
        }
        BlockMove(b2, b, xd, yd, xd + zp-1, yd + zp-1, x * zp, y * zp);
      } else {
        if (yd >= y2) {
          if (yd < x2)
            BlockC(c, x * zp, y * zp, x * zp + zp-1, y * zp + zp-1, kvWhite);
          continue;
        }
        BlockMoveC(c2, c, xd, yd, xd + zp-1, yd + zp-1, x * zp, y * zp, -1);
      }
    }
  if (!fColor)
    DeallocateBitmap(bT);
  return fTrue;
}


// Replace the contents of a color bitmap with the pixels within an arbitrary
// quadrilateral in another bitmap. The source quadrilateral pixels get
// stretched to fit within the rectangle bounds of the destination bitmap.

void ColmapStretchToEdge(colmap c1, constcolmap c2, int x1, int y1,
  int x2, int y2, int x3, int y3, int x4, int y4)
{
  int xmax = CX(c1), ymax = CY(c1), x, y, m1, n1, m2, n2;

  if (FNull(c1) || FNull(c2))
    return;
  Legalize(c2, &x1, &y1);
  Legalize(c2, &x2, &y2);
  Legalize(c2, &x3, &y3);
  Legalize(c2, &x4, &y4);
  for (y = 0; y < ymax; y++) {

    // Compute the source coordinates in the quadrilateral that map to the
    // current row in the destination.
    m1 = x1 + y * (x3 - x1) / ymax;
    n1 = y1 + y * (y3 - y1) / ymax;
    m2 = x2 + y * (x4 - x2) / ymax;
    n2 = y2 + y * (y4 - y2) / ymax;

    for (x = 0; x < xmax; x++)
      CP(c1, x, y,
        CG(c2, m1 + x * (m2 - m1) / xmax, n1 + x * (n2 - n1) / xmax));
  }
}


// Display information about the pixels in a color bitmap, such as number and
// percentage of pure black pixels, white pixels, and grayscale pixels, along
// with the amount of red, green, and blue in the average pixel.

long ColmapCount(constcolmap c)
{
  char sz[cchSzMax], sz1[cchSzMax], sz2[cchSzMax], sz3[cchSzMax];
  long tot, bl = 0, wh = 0, gr = 0, r = 0, g = 0, b = 0, l;
  int x, y;
  KV kv;

  tot = (long)CX(c)*CY(c);
  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++) {
      kv = CG(c, x, y);
      if (kv == kvBlack)
        bl++;
      else if (kv == kvWhite)
        wh++;
      else if (RgbR(kv) == RgbG(kv) && RgbG(kv) == RgbB(kv))
        gr++;
      r += RgbR(kv); g += RgbG(kv); b += RgbB(kv);
    }
  sprintf(S(sz1),
    "Total pixels: %ld\nBlack: %ld (%.2f%%), White: %ld (%.2f%%)",
    tot, bl, tot ? (real)bl/(real)tot*100.0 : 0.0,
    wh, tot ? (real)wh/(real)tot*100.0 : 0.0);
  l = tot-bl-wh-gr;
  sprintf(S(sz2), ", Grayscale: %ld (%.2f%%), Color: %ld (%.2f%%)\n",
    gr, tot ? (real)gr/(real)tot*100.0 : 0.0,
    l, tot ? (real)l/(real)tot*100.0 : 0.0);
  l = r + g + b;
  sprintf(S(sz3), "Red: %.2f%%, Green: %.2f%%, Blue: %.2f%%, Total: %.2f%%\n",
    tot ? (real)r/255.0/(real)tot*100.0 : 0.0,
    tot ? (real)g/255.0/(real)tot*100.0 : 0.0,
    tot ? (real)b/255.0/(real)tot*100.0 : 0.0,
    tot ? (real)l/765.0/(real)tot*100.0 : 0.0);
  sprintf(S(sz), "%s%s%s", sz1, sz2, sz3);
  PrintSz_N(sz);
  return wh;
}


// Search a color bitmap for a pixel that has the specified characteristic.

long ColmapFind(constcolmap c, int nOp)
{
  int x, y, n, i;
  byte *pb;
  KV kvMax, kv;

  switch (nOp) {

  // Return the maximum value pixel in a color bitmap.
  case 0:
    kvMax = 0;
    for (y = 0; y < CY(c); y++) {
      pb = CJ(c, CI(c, 0, y));
      for (x = 0; x < CX(c); x++) {
        kv = CK(pb);
        if (kv > kvMax)
          kvMax = kv;
      }
    }
    return kvMax;

  // Return a random color pixel that's adjacent to a black pixel.
  case 1:
    for (i = 1; i <= DIRS; i++) {
      n = 0;
      for (y = 1; y < CY(c)-1; y++)
        for (x = 1; x < CX(c)-1; x++)
          if (CG(c, x, y) > kvBlack &&
            ((CG(c, x, y-1) == kvBlack) + (CG(c, x-1, y) == kvBlack) +
            (CG(c, x, y+1) == kvBlack) + (CG(c, x+1, y) == kvBlack) == i))
            n++;
      if (n < 1)
        continue;
      n = Rnd(1, n);
      for (y = 1; y < CY(c)-1; y++)
        for (x = 1; x < CX(c)-1; x++)
          if (CG(c, x, y) > kvBlack &&
            ((CG(c, x, y-1) == kvBlack) + (CG(c, x-1, y) == kvBlack) +
            (CG(c, x, y+1) == kvBlack) + (CG(c, x+1, y) == kvBlack) == i)) {
            n--;
            if (n < 1)
              return LFromWW(x, y);
          }
    }
    return LFromWW(CX(c) >> 1, CY(c) >> 1);

  default:
    Assert(fFalse);
  }
  return 0;
}


// Zoom a color bitmap by 200% plus one extra pixel. Boundaries or edges
// between different colored pixels in the old bitmap, become lines of pixels
// in the new whose color is the difference between those colors.

flag FColmapAccentBoundary(colmap *pc)
{
  bitmap c = *pc, c2;
  int x, y, xnew, ynew;
  KV kv0, kv1, kv2, kv3, kvT;

  if (!FAllocateColmap(&c2, (CX(c) << 1) + 1, (CY(c) << 1) + 1))
    return fFalse;
  ColmapSet(c2, kvBlack);
  for (y = -1; y < CY(c); y++) {
    ynew = (y + 1) << 1;
    for (x = -1; x < CX(c); x++) {
      xnew = (x + 1) << 1;
      kv0 = Cg(c, x, y);   kv1 = Cg(c, x+1, y);
      kv2 = Cg(c, x, y+1); kv3 = Cg(c, x+1, y+1);
      if (kv0 != kv3 || kv1 != kv3 || kv2 != kv3) {

        // When 3+ colors meet at a corner, pick the brightest difference.
        kvT = KvDiff(kv0, kv3);
        kvT = Max(kvT, KvDiff(kv1, kv3));
        kvT = Max(kvT, KvDiff(kv2, kv3));
        kvT = Max(kvT, KvDiff(kv1, kv2));
        CP(c2, xnew, ynew, kvT);
      }
      if (kv1 != kv3)
        CP(c2, xnew+1, ynew, KvDiff(kv1, kv3));
      if (kv2 != kv3)
        CP(c2, xnew, ynew+1, KvDiff(kv2, kv3));
    }
  }
  DeallocateColmap(c);
  *pc = c2;
  return fTrue;
}


// Change a color bitmap so that each pixel's RGB components are set to the
// maximum difference between that pixel and all its neighbors.

flag FColmapAccentContrast(colmap c, flag fCorner)
{
  colmap c2 = NULL;
  int x, y, d, dMax = DIRS + fCorner*DIRS, xT, yT,
    nR, nG, nB, nRT, nGT, nBT, nRHi, nGHi, nBHi;
  KV kv, kvT;

  if (!FColmapCopy(c, &c2))
    return fFalse;
  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++) {
      nRHi = nGHi = nBHi = 0;
      kv = CG(c2, x, y);
      nR = RgbR(kv); nG = RgbG(kv); nB = RgbB(kv);
      for (d = 0; d < dMax; d++) {
        xT = x + xoff[d]; yT = y + yoff[d];
        if (FLegalC(c, xT, yT)) {
          kvT = Cg(c2, xT, yT);
          nRT = RgbR(kvT); nGT = RgbG(kvT); nBT = RgbB(kvT);
          nRT = (nR - nRT); nGT = (nG - nGT); nBT = (nB - nBT);
          if (nRT > nRHi)
            nRHi = nRT;
          if (nGT > nGHi)
            nGHi = nGT;
          if (nBT > nBHi)
            nBHi = nBT;
        }
      }
      CP(c, x, y, Rgb(nRHi, nGHi, nBHi));
    }
  DeallocateColmap(c2);
  return fTrue;
}


// Offset the content in a color bitmap by the given number of pixels. This
// doesn't change the bitmap's size or lose any data, where pixels that move
// off an edge of the bitmap reappear on the opposite.

flag FColmapSlide(colmap c, int dx, int dy)
{
  colmap c2 = NULL;

  if (!FColmapCopy(c, &c2))
    return fFalse;
  while (dx < 0)
    dx += CX(c);
  while (dx >= CX(c))
    dx -= CX(c);
  while (dy < 0)
    dy += CY(c);
  while (dy >= CY(c))
    dy -= CY(c);

  // Copy content in four quadrants from the old color bitmap to the new.
  BlockMoveC(c2, c, dx, dy, CX(c)-1, CY(c)-1, 0, 0, -1);
  if (dx > 0)
    BlockMoveC(c2, c, 0, dy, dx-1, CY(c)-1, CX(c)-dx, 0, -1);
  if (dy > 0)
    BlockMoveC(c2, c, dx, 0, CX(c)-1, dy-1, 0, CY(c)-dy, -1);
  if (dx > 0 && dy > 0)
    BlockMoveC(c2, c, 0, 0, dx-1, dy-1, CX(c)-dx, CY(c)-dy, -1);
  DeallocateColmap(c2);
  return fTrue;
}


// Duplicate or delete a range of rows or columns in a color bitmap, shifting
// the rest of the content appropriately.

flag FColmapEditRowColumn(colmap *pc, int z1, int z2, flag fX, flag fAdd)
{
  colmap c = *pc, c2;
  int x, y, d;

  // Do nothing if either index is outside the bounds of the bitmap.
  if (z1 < 0 || (fX && z1 >= CX(c)) || (!fX && z1 >= CY(c)) ||
    z2 < 0 || (fX && z2 >= CX(c)) || (!fX && z2 >= CY(c)))
    return fFalse;
  SortN(&z1, &z2);

  // Figure out the new size of the bitmap.
  d = z2 - z1 + 1;
  x = CX(c); y = CY(c);
  if (fX)
    x += fAdd ? d : -d;
  else
    y += fAdd ? d : -d;
  if (!FAllocateColmap(&c2, x, y))
    return fFalse;

  // Copy content in two halves from the old bitmap to the new.
  if (fX) {
    BlockMoveC(c, c2, 0, 0, fAdd ? z2 : z1-1, CY(c)-1, 0, 0, -1);
    BlockMoveC(c, c2, fAdd ? z1 : z2+1, 0, CX(c)-1, CY(c)-1,
      fAdd ? z2+1 : z1, 0, -1);
  } else {
    BlockMoveC(c, c2, 0, 0, CX(c)-1, fAdd ? z2 : z1-1, 0, 0, -1);
    BlockMoveC(c, c2, 0, fAdd ? z1 : z2+1, CX(c)-1, CY(c)-1,
      0, fAdd ? z2+1 : z1, -1);
  }
  DeallocateColmap(c);
  *pc = c2;
  return fTrue;
}


CONST int rggrfSeen[9] = {NBinary(1,1,0,0,0,0,0,1), NBinary(1,1,0,0,0,1,1,1),
  NBinary(0,0,0,0,0,1,1,1), NBinary(1,1,1,1,0,0,0,1), 0,
  NBinary(0,0,0,1,1,1,1,1), NBinary(0,1,1,1,0,0,0,0),
  NBinary(0,1,1,1,1,1,0,0), NBinary(0,0,0,1,1,1,0,0)};

// Highlight all pixels visible via line of sight, within a certain range from
// a starting point, with colors from another bitmap. Implements the DrawSeen
// operation.

void DrawSeen(colmap c, constbitmap b, constcolmap c2, int x0, int y0, int dr,
  flag fCorner)
{
  int m1, n1, m2, n2, m, n, rSq, xnew, ynew;
  flag fCircle = dr < 0, fAllLegal = FLegalC(c, x0, y0);
  int x, y, dx, dy, xInc, yInc, xInc2, yInc2, d, dInc, z, zMax,
    dm, dn, igrf, e;

  if (fCircle) {
    neg(dr);
    rSq = Sq(dr);
  }
  m1 = NMax(x0 - dr, 0); n1 = NMax(y0 - dr, 0);
  m2 = NMin(x0 + dr, CX(c) - 1); n2 = NMin(y0 + dr, CY(c) - 1);

  // Loop over each potentially visible pixel within the maximum range.
  for (n = n1; n <= n2; n++)
    for (m = m1; m <= m2; m++) {
      if (fCircle && Sq(m - x0) + Sq(n - y0) > rSq)
        continue;
      if (Mg(b, m, n) || CG(c, m, n) == CG(c2, m, n))
        continue;
      dm = NSgn(m - x0); dn = NSgn(n - y0);
      igrf = dn * 3 + dm + 4;

      // For each pixel, check for walls along a line from the viewer to it.
      x = x0; y = y0;
      dx = m - x0; dy = n - y0;
      if (NAbs(dx) >= NAbs(dy)) {
        xInc = NSgn(dx); yInc = 0;
        xInc2 = 0; yInc2 = NSgn(dy);
        zMax = NAbs(dx); dInc = NAbs(dy);
        d = zMax - (!FOdd(dx) && x0 > m);
      } else {
        xInc = 0; yInc = NSgn(dy);
        xInc2 = NSgn(dx); yInc2 = 0;
        zMax = NAbs(dy); dInc = NAbs(dx);
        d = zMax - (!FOdd(dy) && y0 > n);
      }
      d >>= 1;

      // Loop over long axis, adjusting short axis for slope as needed.
      for (z = 0; z <= zMax; z++) {
        if (fAllLegal || (FBetween(x, -1, CX(c)) && FBetween(y, -1, CY(c)))) {
          if (MgFast(b, x, y))
            break;
          if (FLegalC(c, x, y))
            CP(c, x, y, CG(c2, x, y));

          // Set wall pixels adjacent to this space.
          for (e = 0; e < DIRS2; e++) {
            if ((rggrfSeen[igrf] & (1 << e)) == 0)
              continue;
            xnew = x + xoff[e]; ynew = y + yoff[e];
            if (!FBetween(xnew, m1, m2) || !FBetween(ynew, n1, n2) ||
              (fCircle && Sq(xnew - x0) + Sq(ynew - y0) > rSq))
              continue;
            if (MgFast(b, xnew, ynew))
              CP(c, xnew, ynew, CG(c2, xnew, ynew));
          }
        }
        x += xInc; y += yInc; d += dInc;
        if (d >= zMax) {
          x += xInc2; y += yInc2; d -= zMax;

          // Don't let line hop diagonally between set pixels.
          if (!fCorner && MgFast(b, x - dm, y) && MgFast(b, x, y - dn))
            break;
        }
      }
    }
}


// Create a simple random dot stereogram in a color bitmap, based on the two
// level image in a monochrome bitmap. A second color bitmap may be used for
// the random dots.

flag FColmapStereogram(colmap *pc, constbitmap b, constcolmap cRandom,
  int width, int height)
{
  colmap c;
  int x, y, h, w, wDrop = NAbs(height);
  flag fDeep = height < 0, fTexture, f;

  if (!FColmapSizeSet(pc, MX(b), MY(b)))
    return fFalse;
  c = *pc;
  fTexture = cRandom != NULL &&
    CX(cRandom) == CX(c) && CY(cRandom) == CY(c);

  for (y = 0; y < MY(b); y++) {
    for (x = 0; x < MX(b); x++) {
      f = MGCore(b, x, y);
      h = f ? height : 0;
      w = width - h;
      if (x - w < 0 || (f == fDeep && MG(b, x - wDrop, y) == !fDeep))
        CP(c, x, y, fTexture ? CG(cRandom, x, y) : kvRandomRainbow);
      else
        CP(c, x, y, CG(c, x - w, y));
    }
  }
  return fTrue;
}


// Treat the color bitmap as a board of the "Life" cellular automaton, and
// create the next generation. A second bitmap contains indexes of the colors
// to use for each cell in the main bitmap.

long ColmapLifeGenerate(colmap c, colmap *pc2, flag fTorus)
{
  colmap c2, cT = NULL;
  int rgk[8], x, y, i, n, k, xT, yT;
  byte *pb;
  long count = 0L, cbRow, l;

  // If the second bitmap doesn't exist in the right size, it can't contain
  // valid color information, so recreate it with random colors.
  if (*pc2 == NULL || CX(*pc2) != CX(c)+2 || CY(*pc2) != CY(c)+2) {
    if (!FColmapSizeSet(pc2, CX(c)+2, CY(c)+2))
      return -1;
    c2 = *pc2;
    ColmapSet(c2, kvBlack);
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        if (CG(c, x, y) > kvBlack) {
          k = Rnd(0, nHueMax-1);
          CP(c2, x+1, y+1, Rgb(k & 255, k >> 8, 1));
          CP(c, x, y, Hue(k));
        }
    UpdateDisplay();
  } else
    c2 = *pc2;

  // For torus worlds, copy the edge cells to the opposite sides. The extra
  // pixel width around the edges allows quick counting of neighboring cells,
  // without having to check for the edge of the bitmap.
  if (fTorus) {
    for (y = 1; y <= CY(c2)-2; y++) {
      CP(c2, 0, y, CG(c2, CX(c2)-2, y));
      CP(c2, CX(c2)-1, y, CG(c2, 1, y));
    }
    for (x = 1; x <= CX(c2)-2; x++) {
      CP(c2, x, 0, CG(c2, x, CY(c2)-2));
      CP(c2, x, CY(c2)-1, CG(c2, x, 1));
    }
    CP(c2, 0,        0,        CG(c2, CX(c2)-2, CY(c2)-2));
    CP(c2, CX(c2)-1, 0,        CG(c2, 1,        CY(c2)-2));
    CP(c2, 0,        CY(c2)-1, CG(c2, CX(c2)-2, 1       ));
    CP(c2, CX(c2)-1, CY(c2)-1, CG(c2, 1,        1       ));
  }

  if (!FColmapCopy(c2, &cT))
    return -1;
  cbRow = c2->clRow << 2;
  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++) {

      // Quickly count the number of neighboring live cells.
      l = CI(cT, x+1, y+1);
      pb = CJ(cT, l);
      k = (*(pb+1) << 8) | *(pb+2);
      n = *(pb-cbRow) + *(pb+cbRow) + *(pb-3) + *(pb+3) +
        *(pb-cbRow-3) + *(pb-cbRow+3) + *(pb+cbRow-3) + *(pb+cbRow+3);
      if (*pb) {
        if (gs.grfLifeDie & (1 << n)) {

          // A cell dies.
          pb = CJ(c2, l);
          CL(pb, 0, 0, 0);
          CP(c, x, y, kvBlack);
          count++;
        }
      } else {
        if (gs.grfLifeBorn & (1 << n)) {

          // A new cell is born.
          n = 0;
          for (i = 0; i < DIRS2; i++) {
            xT = x + xoff[i]; yT = y + yoff[i];
            pb = CJ(cT, CI(cT, xT+1, yT+1));
            if (*pb)
              rgk[n++] = (*(pb+1) << 8) | *(pb+2);
          }

          // Find the midpoint of three colors of the rainbow of the parent
          // cells, and make a new color of the rainbow based on them.
          k = (rgk[0] + rgk[1] + rgk[2] + 1) / 3;
          if (NAbs(k - rgk[0]) + NAbs(k - rgk[1]) +
            NAbs(k - rgk[2]) > nHue23) {
            k += nHue13 * ((rgk[0] >= nHueHalf) + (rgk[1] >= nHueHalf) +
              (rgk[2] >= nHueHalf) >= 2 ? 1 : -1);
            if (k < 0)
              k += nHueMax;
            else if (k >= nHueMax)
              k -= nHueMax;
          }
          pb = CJ(c2, l);
          CL(pb, k & 255, k >> 8, 1);
          CP(c, x, y, Hue(k));
          count++;
        }
      }
    }
  DeallocateColmap(cT);
  return count;
}


// Treat the color bitmap as a board of the "Evolution" cellular automaton,
// and create the next generation. A second bitmap contains indexes of the
// colors to use for each cell in the main bitmap.

long Evolution(colmap *pc, colmap *pc2, int col, flag fContinue, flag fTorus)
{
  colmap c = *pc, c2, cT = NULL;
  KV rgkv[32];
  byte *pb;
  int x, y, i;
  long count = 0L, cbRow, l;

  if (!FColmapSizeSet(pc2, CX(c)+2, CY(c)+2))
    return -1;
  c2 = *pc2;
  Assert(col < 32);
  for (i = 0; i < col; i++)
    rgkv[i] = Hue(NMultDiv(i, nHueMax, col));

  // For the initial generation, give every pixel a random color.
  if (!fContinue) {
    ColmapSet(c2, kvWhite);
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++) {
        i = Rnd(0, col-1);
        CP(c2, x+1, y+1, i);
        CP(c, x, y, rgkv[i]);
      }
    return 0;
  }

  // For torus worlds, copy the edge cells to the opposite sides. The extra
  // pixel width around the edges allows quick checking of four neighboring
  // cells, without having to check for the edge of the bitmap.
  if (fTorus) {
    for (y = 1; y <= CY(c2)-2; y++) {
      CP(c2, 0, y, CG(c2, CX(c2)-2, y));
      CP(c2, CX(c2)-1, y, CG(c2, 1, y));
    }
    for (x = 1; x <= CX(c2)-2; x++) {
      CP(c2, x, 0, CG(c2, x, CY(c2)-2));
      CP(c2, x, CY(c2)-1, CG(c2, x, 1));
    }
  } else {
    // For non-torus worlds, copy the edge cells to the same sides, to prevent
    // any color flowing to/from off the edges of the bitmap.
    for (y = 1; y <= CY(c2)-2; y++) {
      CP(c2, 0, y, CG(c2, 1, y));
      CP(c2, CX(c2)-1, y, CG(c2, CX(c2)-2, y));
    }
    for (x = 1; x <= CX(c2)-2; x++) {
      CP(c2, x, 0, CG(c2, x, 1));
      CP(c2, x, CY(c2)-1, CG(c2, x, CY(c2)-2));
    }
  }

  if (!FColmapCopy(c2, &cT))
    return -1;
  cbRow = c2->clRow << 2;
  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++) {

      // Quickly check if a neighboring cell is one color higher.
      l = CI(cT, x+1, y+1);
      pb = CJ(cT, l) + 2;
      i = *pb + 1;
      if (i >= col)
        i = 0;
      if (*(pb-cbRow) == i || *(pb-3) == i ||
        *(pb+cbRow) == i || *(pb+3) == i) {
        count++;

        // Set the current cell's color to the neighbor cell's color.
        *(CJ(c2, l) + 2) = i;
        CP(c, x, y, rgkv[i]);
      }
    }
  DeallocateColmap(cT);
  return count;
}


#define cMandy 1536

// Create a Mandelbrot set fractal in a color bitmap, looking upon the
// specified rectangle of coordinates.

void Mandelbrot(colmap c, real xmin, real ymin, real xmax, real ymax,
  int nLimit, flag fShip)
{
  KV rgkv[cMandy], kv;
  byte *pb;
  real x, y, xInc, yInc, m, n, z;
  int iMax = Min(nLimit, cMandy), xa, ya, color, i;

  // Define the set of colors to use for the various depths.
  if (!cs.fGraphNumber) {
    for (i = iMax-1; i >= 0; i--)
      rgkv[i] = Hue(NMultDiv(i, nHueMax, cMandy));
  } else {
    for (i = iMax-1; i >= 0; i--)
      rgkv[i] = iMax - i;
  }

  ColmapSet(c, kvBlack);
  xInc = (xmax - xmin) / CX(c);
  yInc = (ymax - ymin) / CY(c);
  for (ya = 0, y = ymin; ya < CY(c); ya++, y += yInc) {
    pb = CJ(c, CI(c, 0, ya));
    for (xa = 0, x = xmin; xa < CX(c); xa++, x += xInc) {
      m = x; n = y;
      if (!fShip) {
        for (color = 0; color < nLimit; color++) {
          z = m * n;
          m *= m; n *= n;
          if (m + n > 4.0)
            break;
          m = m - n + x;
          n = z + z + y;
        }
      } else {
        for (color = 0; color < nLimit; color++) {
          z = m * n;
          m *= m; n *= n;
          if (m + n > 4.0)
            break;
          m = m - n + x;
          n = z + z + y;
          if (m < 0)
            neg(m);
          if (n < 0)
            neg(n);
        }
      }

      // If the depth limit wasn't reached, set the pixel to a color.
      if (color < nLimit) {
        while (color >= cMandy)
          color -= cMandy;
        kv = rgkv[color];
        CL(pb, RgbR(kv), RgbG(kv), RgbB(kv));
      }
      pb += cbPixelC;
    }
  }
}


/*
******************************************************************************
** Color Bitmap IO Routines
******************************************************************************
*/

// Copy a monochrome bitmap to a color bitmap. Off and on pixels in the
// monochrome bitmap are mapped to the given colors.

flag FColmapGetFromBitmap(colmap *pc, constbitmap b, KV kv0, KV kv1)
{
  colmap c;
  int x, y, nR, nG, nB;
  byte *pb;

  if (!FColmapSizeSet(pc, MX(b), MY(b)))
    return fFalse;
  c = *pc;
  ColmapSet(c, kv0);
  nR = RgbR(kv1); nG = RgbG(kv1); nB = RgbB(kv1);
  for (y = 0; y < CY(c); y++) {
    pb = CJ(c, CI(c, 0, y));
    for (x = 0; x < CX(c); x++) {
      if (MG(b, x, y)) {
        CL(pb, nR, nG, nB);
      }
      pb += cbPixelC;
    }
  }
  CopyClient(b, c);
  return fTrue;
}


// Copy a color bitmap to a monochrome bitmap, in a specified fashion.

flag FColmapPutToBitmap(constcolmap c, bitmap *pb, int nOp)
{
  bitmap b;
  int x, y;
  KV kv;

  if (!FBitmapSizeSet(pb, CX(c), CY(c)))
    return fFalse;
  b = *pb;
  BitmapOff(b);
  switch (nOp) {

  // Non-black pixels in the color bitmap become white, all others black.
  case 0:
    for (y = 0; y < MY(b); y++)
      for (x = 0; x < MX(b); x++)
        if (CG(c, x, y) > 0)
          M1(b, x, y);
    break;

  // Light colors in the color bitmap become white, dark colors become black.
  case 1:
    for (y = 0; y < MY(b); y++)
      for (x = 0; x < MX(b); x++) {
        kv = CG(c, x, y);
        if (RgbR(kv) + RgbG(kv) + RgbB(kv) > 127*3+1)
          M1(b, x, y);
      }
    break;

  // Dither: Colors become white or black in proportion to their brightness.
  case 2:
    for (y = 0; y < MY(b); y++)
      for (x = 0; x < MX(b); x++) {
        kv = CG(c, x, y);
        if (FDither((RgbR(kv) + RgbG(kv) + RgbB(kv)) / 45, x, y))
          M1(b, x, y);
      }
    break;

  default:
    Assert(fFalse);
  }
  CopyClient(c, b);
  return fTrue;
}


// Combine a monochrome bitmap with a color bitmap using the specified
// operator, where off and on pixels in the monochrome bitmap are treated as
// the given colors.

void ColmapOrAndFromBitmap(colmap c, constbitmap b, KV kv0, KV kv1, int nOp)
{
  int x, y, x1, y1;
  KV kv;

  x1 = Min(CX(c), MX(b)); y1 = Min(CY(c), MY(b));
  for (y = 0; y < y1; y++)
    for (x = 0; x < x1; x++) {
      kv = MG(b, x, y) ? kv1 : kv0;
      switch (nOp) {
      case 0: CP(c, x, y, CG(c, x, y) & kv); break;
      case 1: CP(c, x, y, CG(c, x, y) | kv); break;
      case 2: CP(c, x, y, CG(c, x, y) ^ kv); break;
      default: Assert(fFalse);
      }
    }
}


// Blend three monochrome bitmaps together into one color bitmap. Each pixel
// in the color bitmap will be one of eight colors.

flag FColmapBlendFromBitmap(colmap *pc, constbitmap b1, constbitmap b2,
  constbitmap b3)
{
  colmap c;
  int x, y;

  if (b2 == NULL)
    b2 = b1;
  if (b3 == NULL)
    b3 = b2;
  x = Min(MX(b1), MX(b2)); x = Min(x, MX(b3));
  y = Min(MY(b1), MY(b2)); y = Min(y, MY(b3));
  if (!FColmapSizeSet(pc, x, y))
    return fFalse;
  c = *pc;

  // Bitmap #1 is treated as red on black, bitmap #2 is green on black, and
  // bitmap #3 is blue on black, where these RGB values are merged together.
  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++)
      CP(c, x, y, Rgb(MG(b1, x, y)*255, MG(b2, x, y)*255, MG(b3, x, y)*255));
  return fTrue;
}


// Copy a monochrome bitmap to a color bitmap, where off and on pixels in the
// monochrome bitmap are mapped to the passed in colors. All off pixels
// reachable from a start point or points will be filled with a blend of
// colors indicating their distance from the nearest start point.

int ColmapGraphDistance(colmap *pc, constbitmap b, constbitmap b2,
  KV kv0, KV kv1, int x, int y, flag fCorner)
{
  colmap c;
  BFSS *bfss;
  int count = 0, xnew, ynew, d, j = DIRS + fCorner*DIRS;
  long iLo = 0, iHi, iMax = 0, i;
  flag f;

  if ((!FLegal(b, x, y) || MG(b, x, y)) && !FBitmapFind(b, &x, &y, fOff)) {
    PrintSz_W("There are no open sections to graph.\n");
    return -2;
  }
  if (!FColmapGetFromBitmap(pc, b, kv0, kv1))
    return -1;
  c = *pc;
  bfss = RgAllocate(CX(c)*CY(c), BFSS);
  if (bfss == NULL)
    return -1;
  CP(c, x, y, kv1);
  BfssPush(iMax, x, y, 0);

  // Potentially allow all off pixels in a 2nd monochrome bitmap to act as
  // start points to flood from, if it exists and is same size as main bitmap.
  f = fFalse;
  if (b2 != NULL && MX(b2) == MX(b) && MY(b2) == MY(b))
    for (ynew = 0; ynew < MY(b2); ynew++)
      for (xnew = 0; xnew < MX(b2); xnew++)
        if (!MG(b2, xnew, ynew)) {
          // Reject 2nd bitmap if it contains an off pixel not off in main.
          if (MG(b, xnew, ynew))
            goto LBegin;
        } else if (!MG(b, xnew, ynew))
          f = fTrue;
  // Accept 2nd bitmap if it contains at least one on pixel off in the main.
  if (f)
    for (ynew = 0; ynew < MY(b2); ynew++)
      for (xnew = 0; xnew < MX(b2); xnew++)
        if (!MG(b2, xnew, ynew)) {
          Assert(!MG(b, xnew, ynew));
          CP(c, xnew, ynew, kv1);
          BfssPush(iMax, xnew, ynew, 0);
        }

LBegin:
  // Flood bitmap from start points, marking distances for each pixel touched.
  iHi = iMax;
  while (iLo < iHi) {
    count++;
    for (i = iLo; i < iHi; i++) {
      x = bfss[i].x; y = bfss[i].y;
      for (d = 0; d < j; d++) {
        xnew = x + xoff[d]; ynew = y + yoff[d];
        if (FLegalC(c, xnew, ynew) && CG(c, xnew, ynew) != kv1) {
          CP(c, xnew, ynew, kv1);
          BfssPush(iMax, xnew, ynew, count);
        }
      }
    }
    iLo = iHi; iHi = iMax;
  }

  // Map distance numbers to colors.
  if (!cs.fGraphNumber) {
    for (i = 0; i < iMax; i++)
      CP(c, bfss[i].x, bfss[i].y,
        Hue(NMultDiv(bfss[i].parent, nHueMax, count)));
  } else {
    for (i = 0; i < iMax; i++)
      CP(c, bfss[i].x, bfss[i].y, bfss[i].parent);
  }
  DeallocateP(bfss);
  return count - 1;
}


#define FValidGraph(x, y) (Mg(b, x, y) && Cg(c, x, y) == kv1)

// Copy a monochrome bitmap to a color bitmap, where off and on pixels in the
// monochrome bitmap are mapped to the passed in colors. One pixel wide walls
// that don't surround inaccessible sections (and aren't linked on both ends
// to walls surrounding them) will be set to blend of colors indicating their
// distance from the farthest wall endpoint within the local wall network. In
// other words, for each wall segment this determines its "importance", or
// graphs the approximate distance it takes to walk from one side of the wall
// to the other (if even possible) by going around the end of the wall.

long ColmapGraphDistance2(colmap *pc, constbitmap b, KV kv0, KV kv1,
  flag fCorner)
{
  colmap c;
  PT *rgpt;
  int mpgrff[256], x, y, xnew, ynew, d, dSav, cd, grf, i;
  long cpt = 0, ipt, iMax = 0;
  KV kv;

  if (!FColmapGetFromBitmap(pc, b, kv0, kv1))
    return -1;
  c = *pc;
  rgpt = RgAllocate(CX(c)*CY(c), PT);
  if (rgpt == NULL)
    return -1;
  if (fCorner) {
    for (i = 0; i < 256; i++)
      mpgrff[i] = fFalse;
    for (i = 0; i < 25; i++)
      mpgrff[rggrfNeighbor[i]] = fTrue;
  }

  // Determine the set of wall endpoints to start flooding from.
  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x++)
      if (MG(b, x, y)) {
        if (!fCorner) {

          // Starts for orthogonal graphing are on pixels with <= 1 neighbor.
          cd = 0;
          for (d = 0; d < DIRS; d++)
            if (Mg(b, x + xoff[d], y + yoff[d])) {
              if (cd > 0)
                goto LNext;
              cd = 1;
            }
        } else {

          // Starts for diagonal graphing are on pixels with no more than one
          // consecutive group of <= 3 neighbors.
          grf = 0;
          for (d = 0; d < DIRS2; d++) {
            grf <<= 1;
            if (Mg(b, x + xoff[d], y + yoff[d]))
              grf |= 1;
          }
          if (!mpgrff[grf])
            goto LNext;
        }
        rgpt[cpt].x = x; rgpt[cpt].y = y;
        cpt++;
LNext:
        ;
      }

  // Flood the bitmap from the start points, marking distances for each pixel.
  while (cpt > 0) {
    iMax++;
    for (ipt = 0; ipt < cpt; ipt++) {
      x = rgpt[ipt].x; y = rgpt[ipt].y;
      if (CG(c, x, y) != kv1)
        goto LStop;
      cd = 0;
      if (!fCorner) {

        // For orthogonal graphing, stop flooding along a wall if the current
        // pixel has > 1 unflooded neighbor.
        for (d = 0; d < DIRS; d++) {
          xnew = x + xoff[d]; ynew = y + yoff[d];
          if (FValidGraph(xnew, ynew)) {
            if (cd > 0)
              goto LStop;
            cd = 1;
            dSav = d;
          }
        }
      } else {

        // For diagonal graphing, first get the count and arrangement of the
        // current cell's unflooded neighbors.
        grf = 0;
        for (d = 0; d < DIRS2; d++) {
          grf <<= 1;
          xnew = x + xoff[d]; ynew = y + yoff[d];
          if (FValidGraph(xnew, ynew)) {
            grf |= 1;
            cd++;
            dSav = d;
          }
        }

        // If current pixel has no more than one consecutive group of <= 3
        // neighbors, pick a direction to flood in that will keep that true.
        if (cd > 1) {
          if (mpgrff[grf])
            CP(c, x, y, iMax);
          for (d = 0; d < DIRS2; d++) {
            xnew = x + xoff[d]; ynew = y + yoff[d];
            if (FValidGraph(xnew, ynew)) {
              grf = 0;
              for (i = 0; i < DIRS2; i++) {
                grf <<= 1;
                if (FValidGraph(xnew + xoff[i], ynew + yoff[i])) {
                  grf |= 1;
                  cd++;
                  dSav = d;
                }
              }
              if (cd >= 1 && mpgrff[grf])
                goto LGo;
            }
          }
          goto LStop;
        }
      }
      CP(c, x, y, iMax);
      if (cd < 1)
        goto LStop;
LGo:
      x += xoff[dSav]; y += yoff[dSav];
      rgpt[ipt].x = x; rgpt[ipt].y = y;
      continue;
LStop:
      rgpt[ipt] = rgpt[cpt-1];
      cpt--; ipt--;
    }
  }

  // Map distance numbers to colors.
  if (!cs.fGraphNumber)
    for (y = 0; y < MY(b); y++)
      for (x = 0; x < MX(b); x++)
        if (MG(b, x, y)) {
          kv = CG(c, x, y);
          if (kv != kv1)
            CP(c, x, y, Hue(NMultDiv(kv, nHueMax, iMax)));
        }
  DeallocateP(rgpt);
  return iMax;
}


// Copy a monochrome bitmap to a color bitmap, with antialiasing. Off and on
// pixels in the monochrome bitmap are mapped to the given colors. The color
// bitmap will be zoomed z times smaller, where each pixel will become the
// right ratio between the off and on colors, based on the number of off and
// on pixels in the monochrome bitmap that map to that pixel.

flag FColmapAntialias(colmap *pc, constbitmap b, KV kv0, KV kv1, int z)
{
  colmap c;
  int x, y, x2, y2, xmax, ymax, n, zSq;

  xmax = MX(b) / z; ymax = MY(b) / z;
  if (!FColmapSizeSet(pc, xmax, ymax))
    return fFalse;
  c = *pc;
  zSq = Sq(z);
  for (y = 0; y < ymax; y++)
    for (x = 0; x < xmax; x++) {
      n = 0;
      for (y2 = 0; y2 < z; y2++)
        for (x2 = 0; x2 < z; x2++)
          n += MG(b, x*z + x2, y*z + y2);
      CP(c, x, y, KvBlendN(kv0, kv1, n, zSq));
    }
  return fTrue;
}


// Load a color bitmap from a Windows bitmap format file. It is assumed
// FReadBitmapHeader() has already been called.

flag FReadColmapCore(colmap *pc, FILE *file, int x, int y, int z, int k)
{
  colmap c;
  int cb, i;
  byte *pb, bR, bG, bB, ch;
  KV rgkv[256], kv;

  Assert(z == 4 || z == 8 || z == 16 || z == 24 || z == 32);
  // RgbQuad
  // Data
  if (!FColmapSizeSet(pc, x, y))
    return fFalse;
  c = *pc;

  // Figure out the bytes per row in this color bitmap.
  if (z == 32)
    cb = CX(c) << 2;
  else if (z == 24)
    cb = CX(c) * 3;
  else if (z == 16) {
    for (i = 0; i < 12; i++)
      skipbyte();
    cb = CX(c) << 1;
  } else {
    // Read in the color palette to translate indexes to RGB values.
    Assert(k <= 256);
    if (k == 0)
      k = (z == 8) ? 256 : 16;
    for (i = 0; i < k; i++) {
      bB = getbyte(); bG = getbyte(); bR = getbyte();
      skipbyte();
      rgkv[i] = Rgb(bR, bG, bB);
    }
    if (z == 8)
      cb = CX(c);
    else
      cb = (CX(c) + 1) >> 1;
  }
  // Figure out the number of padding bytes after each row.
  cb = (4 - (cb & 3)) & 3;

  for (y = CY(c)-1; y >= 0; y--) {
    pb = CJ(c, CI(c, 0, y));
    for (x = 0; x < CX(c); x++) {
      if (z >= 24) {
        bB = getbyte(); bG = getbyte(); bR = getbyte();
        if (z == 32)
          skipbyte();
      } else {
        if (z == 16) {
          i = WRead(file);
          kv = Rgb((i >> 11) << 3, (i >> 5 & 63) << 2, (i & 31) << 3);
        } else if (z == 8) {
          ch = getbyte();
          kv = rgkv[ch];
        } else {
          if (!FOdd(x))
            ch = getbyte();
          i = FOdd(x) ? (ch & 15) : (ch >> 4);
          kv = rgkv[i];
        }
        bR = RgbR(kv); bG = RgbG(kv); bB = RgbB(kv);
      }
      CL(pb, bR, bG, bB);
      pb += cbPixelC;
    }
    for (x = 0; x < cb; x++)
      skipbyte();
  }
  return fTrue;
}


// Save a color bitmap to a Windows bitmap format file.

void WriteColmap(constcolmap c, FILE *file)
{
  int x, y, cb;
  dword dw;

  cb = (4 - ((CX(c)*3) & 3)) & 3;
  // BitmapFileHeader
  putbyte('B'); putbyte('M');
  dw = 14+40 + 0 + CY(c)*(CX(c)*3+cb);
  putlong(dw);
  putword(0); putword(0);
  putlong(14+40 + 0);
  // BitmapInfo / BitmapInfoHeader
  putlong(40);
  putlong(CX(c)); putlong(CY(c));
  putword(1); putword(24);
  putlong(0 /*BI_RGB*/); putlong(0);
  putlong(0); putlong(0);
  putlong(0); putlong(0);
  // RgbQuad
  // Data
  for (y = CY(c)-1; y >= 0; y--) {
    for (x = 0; x < CX(c); x++) {
      dw = CG(c, x, y);
      putbyte(RgbB(dw)); putbyte(RgbG(dw)); putbyte(RgbR(dw));
    }
    for (x = 0; x < cb; x++)
      putbyte(0);
  }
}


// Load a color Targa bitmap from a file, into a new color bitmap.

flag FReadColmapTarga(colmap *pc, FILE *file)
{
  colmap c;
  int x, y, i;
  byte ch, chR, chG, chB;

  for (i = 0; i < 12; i++)
    skipbyte();
  fscanf(file, "%c%c", S1(&ch), S1(&chR));
  x = (int)(chR)*256 + (int)ch;
  fscanf(file, "%c%c", S1(&ch), S1(&chR));
  y = (int)(chR)*256 + (int)ch;
  if (!FColmapSizeSet(pc, x, y))
    return fFalse;
  c = *pc;
  skipword();
  for (y = CY(c)-1; y >= 0; y--) {
    for (x = 0; x < CX(c); x++) {
      fscanf(file, "%c%c%c%c", S1(&chB), S1(&chG), S1(&chR), S1(&ch));
      CP(c, x, y, Rgb(chR, chG, chB));
    }
  }
  return fTrue;
}


// Load a color PC Paintbrush bitmap from a file, into a new color bitmap.

flag FReadColmapPaint(colmap *pc, FILE *file)
{
  colmap c;
  int x1, y1, x2, y2, x, y, i;
  byte ch, chR, chG, chB;
  KV rgkv[256];

  // Read the header
  for (i = 0; i < 3; i++)
    skipbyte();
  ch = getbyte();
  if (ch != 8)
    return fFalse;
  x1 = WRead(file); y1 = WRead(file);
  x2 = WRead(file); y2 = WRead(file);
  x = x2 - x1 + 1; y = y2 - y1 + 1;
  for (i = 0; i < 116; i++)
    skipbyte();
  if (!FColmapSizeSet(pc, x, y))
    return fFalse;
  c = *pc;

  // Load the color indexes
  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++) {
      ch = getbyte();
      if (ch < 192)
        i = 0;
      else {
        i = ch - 193;
        ch = getbyte();
      }
      CP(c, x, y, ch);
      while (i > 0) {
        CP(c, ++x, y, ch);
        i--;
      }
    }

  // Load the palette
  skipbyte();
  for (i = 0; i < 256; i++) {
    fscanf(file, "%c%c%c", S1(&chR), S1(&chG), S1(&chB));
    rgkv[i] = Rgb(chR, chG, chB);
  }

  // Apply the palette
  for (y = 0; y < CY(c); y++)
    for (x = 0; x < CX(c); x++) {
      i = RgbR(CG(c, x, y));
      CP(c, x, y, rgkv[i]);
    }
  return fTrue;
}


// Save a color bitmap to a Targa format file.

void WriteColmapTarga(constcolmap c, FILE *file)
{
  int x, y;
  KV kv;

  fprintf(file, "%c%c%c%c%c%c%c%c",
    chNull, chNull, '\2', chNull, chNull, chNull, chNull, chNull);
  x = CX(c); y = CY(c);
  fprintf(file, "%c%c%c%c",
    (char)(x % 256), (char)(x / 256),
    (char)(y % 256), (char)(y / 256));
  fprintf(file, "%c%c%c%c",
    (char)(CX(c) % 256), (char)(CX(c) / 256),
    (char)(CY(c) % 256), (char)(CY(c) / 256));
  fprintf(file, " %c", chNull);
  for (y = CY(c)-1; y >= 0; y--) {
    for (x = 0; x < CX(c); x++) {
      kv = CG(c, x, y);
      fprintf(file, "%c%c%c%c",
        (char)(RgbB(kv)), (char)(RgbG(kv)), (char)(RgbR(kv)), chNull);
    }
  }
}


// Load a Daedalus color bitmap from a file, into a new color bitmap.

flag FReadDaedalusColmapCore(colmap *pc, FILE *file, int x, int y)
{
  colmap c;
  int i, cn, nCur;
  KV kv;
  char ch;

  if (!FColmapSizeSet(pc, x, y))
    return fFalse;
  c = *pc;
  ColmapSet(c, kvBlack);
  skipcrlf();
  for (y = 0; y < CY(c); y++) {
    x = 0;
    ch = getbyte();
    loop {
      if (ch < ' ') {
        skiplf();
        break;
      }
      if (ch < 'a' || ch > 'z') {
        // Single pixels start with '{' or characters before 'a'.
        cn = 1;
        if (ch == '{')
          ch = getbyte();
        else if (ch == '|') {
          // Found the duplicate row marker '|'.
          Assert(x == 0 && y > 0);
          for (nCur = 0, ch = getbyte(); FDigitCh(ch); ch = getbyte())
            nCur = nCur * 10 + (ch - '0');
          BlockMoveC(c, c, 0, y-1, CX(c)-1, y-1, 0, y, -1);
          while (--nCur > 0) {
            y++;
            BlockMoveC(c, c, 0, y-1, CX(c)-1, y-1, 0, y, -1);
          }
          skipcrlf();
          break;
        }
      } else {
        // A run of 2 to 27 pixels is indicated by characters 'a' thru 'z'.
        cn = ch - 'a' + 2;
        ch = getbyte();
      }
      if (ch >= 'a') {
        // One of 27 common colors is encoded by characters 'a' thru '{'.
        i = ch - 'a';
        kv = Rgb((i / 9)*255 >> 1, (i / 3 % 3)*255 >> 1, (i % 3)*255 >> 1);
        ch = getbyte();
      } else {
        // Standard case has four characters encode a 24 bit RGB value.
        kv = 0;
        for (i = 0; i < 4; i++) {
          kv = (kv << 6) + (ch - '!');
          ch = getbyte();
        }
      }
      // Set the current run of pixels in the bitmap.
      while (cn--) {
        CP(c, x, y, kv);
        x++;
      }
    }
  }
  return fTrue;
}


// Save a color bitmap to a Daedalus color bitmap custom format file.

void WriteDaedalusColmap(constcolmap c, FILE *file, flag fClip)
{
  int x, y, xmax, i, cn, nr, ng, nb, cRow = 0;
  KV kv;
  flag fBlank;

  fprintf(file, "DB\n%d %d 24\n\n", CX(c), CY(c));
  for (y = 0; y <= CY(c); y++) {

    // Check if current row same as previous. If so write '|' dup marker.
    if (y > 0) {
      if (y < CY(c)) {
        for (x = 0; x < CX(c); x++)
          if (CG(c, x, y) != CG(c, x, y-1))
            goto LNormal;
        cRow++;
        continue;
      }

LNormal:
      // Output number of times row was duplicated
      if (cRow > 0) {
        if (!fBlank || cRow > 1) {
          putbyte('|');
          if (cRow > 1)
            fprintf(file, "%d", cRow);
        }
        putbyte('\n');
        cRow = 0;
      }
    }
    if (y >= CY(c))
      continue;

    xmax = CX(c) - 1;
    if (fClip)
      while (xmax >= 0 && CG(c, xmax, y) == kvBlack)
        xmax--;
    fBlank = (xmax < 0);
    for (x = 0; x <= xmax; x += cn) {
      kv = CG(c, x, y);

      // Check for runs of consecutive colors in a row.
      for (cn = 1; cn <= 26 && x + cn < CX(c) && CG(c, x + cn, y) == kv; cn++)
        ;
      // Indicate a run of 2 to 27 pixels by characters 'a' thru 'z'.
      if (cn > 1)
        putbyte('a' + cn - 2);

      nr = RgbR(kv); ng = RgbG(kv); nb = RgbB(kv);
      if ((nr == 0 || nr == 255 || nr == 127) &&
        (ng == 0 || ng == 255 || ng == 127) &&
        (nb == 0 || nb == 255 || nb == 127)) {
        // One of 27 common colors is encoded by characters 'a' thru '{'.
        if (cn <= 1)
          putbyte('{');
        putbyte('a' + ((nr+1) >> 7)*9 + ((ng+1) >> 7)*3 + ((nb+1) >> 7));
      } else {
        // Standard case requires four chars to encode a 24 bit RGB value.
        for (i = 0; i < 4; i++)
          putbyte('!' + (int)((kv >> (3-i)*6) & 0x3f));
      }
    }
    putbyte('\n');
  }
}


// Given a color, return a character approximating the color's grayscale
// value. Called from WriteTextColmap().

char ChFromKv(KV kv)
{
  int n;

  if (cs.lGrayscale < 0) {

    // Simple case: Treat RGB channels as the same brightness.
    n = RgbR(kv) + RgbG(kv) + RgbB(kv);
  } else {

    // Advanced case: Each of the RGB channels has its own brightness.
    n  = RgbR(kv) * (cs.lGrayscale / 1000000);
    n += RgbG(kv) * (cs.lGrayscale / 1000 % 1000);
    n += RgbB(kv) * (cs.lGrayscale % 1000);
    n = n * 766 / 255000;
  }
  return rgchAscii[n * cchAscii / 766];
}


// Save a color bitmap to a text file, with each color pixel written as a
// character based on the color's grayscale value.

void WriteTextColmap(constcolmap c, FILE *file, flag fClip)
{
  int x, y, x2 = CX(c)-1;
  char ch;

  for (y = 0; y < CY(c); y++) {
    if (fClip)
      for (x2 = CX(c)-1; x2 >= 0 && ChFromKv(CG(c, x2, y)) == ' '; x2--)
        ;
    for (x = 0; x <= x2; x++) {
      ch = ChFromKv(CG(c, x, y));
      putbyte(ch);
    }
    putbyte('\n');
  }
}


// Write a color value to a string.

void ConvertKvToSz(KV kv, char *sz, int cch)
{
  int i;

  // If the color is one of the standard colors in the color table, use that
  // string, otherwise compose a string with RGB numbers.

  for (i = 0; i < cColor; i++)
    if (kv == rgkv[i]) {
      sprintf(SS(sz, cch), "%s", rgszColor[i]);
      return;
    }
  if (kv < 0) {
    sprintf(SS(sz, cch), "-1");  // Catch invalid colors
    return;
  }
  sprintf(SS(sz, cch), "rgb %d %d %d", RgbR(kv), RgbG(kv), RgbB(kv));
}


// Convert a color stored in a string to an actual color value, advancing the
// string pointer to after the string. This can parse all colors generated
// with ConvertKvToSz. This may call itself recursively to handle functions
// that take colors as parameters, e.g. "blend red orange".

KV ParseColorCore(char **ppch, flag fAllow)
{
  char line[cchSzDef];
  int i;

  SkipSpaces(*ppch);
  sscanf(*ppch, "%s", S(line));
  SkipToSpace(*ppch);

  // Check for standard colors in the color table.
  for (i = 0; i < cColor; i++)
    if (ChCap(line[0]) == ChCap(rgszColor[i][0]) &&
      ChCap(line[1]) == ChCap(rgszColor[i][1]) &&
      ChCap(line[2]) == ChCap(rgszColor[i][2]) && line[3] == chNull)
      return rgkv[i];

  // Check for strings that can translate to multiple colors.
  if (FEqSz(line, "???"))
    return kvRandomRainbow;
  else if (FEqSz(line, "????"))
    return kvRandom;
  else if (FEqSzI(line, "dlg"))
    return KvDialog();

  // Check for special color patterns, if they're allowed.
  if (fAllow) {
    if (FEqSz(line, "O"))
      return -crpRing;
    else if (FEqSz(line, "X"))
      return -crpSpoke;
    else if (FEqSz(line, "*"))
      return -crpFade;
    else if (FEqSz(line, "|"))
      return -crpY;
    else if (FEqSz(line, "-"))
      return -crpX;
    else if (FEqSz(line, "\\"))
      return -crpDiagUL;
    else if (FEqSz(line, "/"))
      return -crpDiagUR;
    else if (FEqSz(line, "#"))
      return -crpSquare;
    else if (FEqSz(line, "+"))
      return -crpDiamond;
    else if (FEqSz(line, "@"))
      return -crpSpiral;
    else if (FEqSz(line, "o"))
      return -crpRing2;
    else if (FEqSz(line, "8"))
      return -crpRings;
    else if (FEqSz(line, "x"))
      return -crpSpoke2;
    else if (FEqSz(line, "?"))
      return -crpRainbow;
    else if (FEqSz(line, "??"))
      return -crpRandom;
    else if (FEqSz(line, "?????"))
      return -crpRandom2;
  }

  // Numbers are treated as that index in the color table.
  if (FDigitCh(line[0])) {
    i = atoi(line);
    if (FBetween(i, 0, cColor-1))
      return rgkv[i];
  }

  // Unknown strings will be parsed via the scripting language.
  return -crpUnknown;
}


// Convert a color stored in a string to an actual color value.

KV ParseColor(char *sz, flag fAllow)
{
  KV kv;
  char *szT = sz;

  kv = ParseColorCore(&szT, fAllow);
  if (kv == -crpUnknown)
    return ParseExpression(sz);
  if (kv < 0 && kv > -crpUnknown)
    cs.nReplacePattern = -kv;
  return kv;
}


// Read a string representing a color from a file, and parse it into a value.

KV ReadColor(FILE *file)
{
  char sz[cchSzDef];
  int i;

  for (i = 0; i < cchSzDef && (sz[i] = getbyte()) >= ' '; i++)
    ;
  sz[i] = chNull;
  return ParseColor(sz, fFalse);
}

/* color.cpp */
