/*
** Daedalus (Version 2.5) File: threed.cpp
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains general graphics routines, that operate on a monochrome
** or color bitmap being treated as a 3D or 4D bitmap, unrelated to Mazes.
**
** Created: 3/12/1990.
** Last code change: 10/10/2013.
*/

#include <stdio.h>
#include <math.h>
#include "util.h"
#include "graphics.h"
#include "color.h"
#include "threed.h"


// Offset table for 3D/4D orthogonal and 2D diagonal directions.
CONST int woff3[DIRS4 + 4] = { 0, 0, 0, 0, 0, 0,-1, 1,  0, 0, 0, 0};
CONST int xoff3[DIRS4 + 4] = { 0,-1, 0, 1, 0, 0, 0, 0, -1,-1, 1, 1};
CONST int yoff3[DIRS4 + 4] = {-1, 0, 1, 0, 0, 0, 0, 0, -1, 1, 1,-1};
CONST int zoff3[DIRS4 + 4] = { 0, 0, 0, 0,-1, 1, 0, 0,  0, 0, 0, 0};


/*
******************************************************************************
** 3D Bitmap Routines
******************************************************************************
*/

// Allocate a 3D bitmap of a given size. The z levels will be arranged in a
// grid with no more than w levels per row.

flag FAllocateCube(cubemap *t, int x, int y, int z, int w)
{
  if (x < 0 || y < 0 || z < 0 ||
    x > xCubeMax || y > yCubeMax || z > zCubeMax) {
    PrintSz_E("Can't create 3D bitmap that large!\n");
    return fFalse;
  }
  if (!FAllocateBitmap(t, x * Min(w, z), y * ((z + w-1) / w), NULL))
    return fFalse;
  TX(*t) = x; TY(*t) = y; TZ(*t) = z; TW(*t) = w;
  return fTrue;
}


// Allocate a color 3D bitmap of a given size.

flag FAllocateCubeC(colmap *c, int x, int y, int z, int w)
{
  if (x < 0 || y < 0 || z < 0 ||
    x > xCubeMax || y > yCubeMax || z > zCubeMax) {
    PrintSz_E("Can't create color 3D bitmap that large!\n");
    return fFalse;
  }
  if (!FAllocateColmap(c, x * Min(w, z), y * ((z + w-1) / w)))
    return fFalse;
  TX(*c) = x; TY(*c) = y; TZ(*c) = z; TW(*c) = w;
  return fTrue;
}


// Make a 3D bitmap be a given size, allocating or reallocating if necessary.

flag FCubeSizeSet(cubemap *t, int xnew, int ynew, int znew, int wnew)
{
  if (!FBitmapSizeSet(t,
    xnew * Min(wnew, znew), ynew * ((znew + wnew-1) / wnew)))
    return fFalse;
  TX(*t) = xnew; TY(*t) = ynew; TZ(*t) = znew; TW(*t) = wnew;
  return fTrue;
}


// Make a 4D bitmap be a given size, allocating or reallocating if necessary.
// A 4D bitmap is a w*z grid of levels, each x*y pixels in size.

flag FTesseractSizeSet(cubemap *t, int wnew, int xnew, int ynew, int znew)
{
  if (!FBitmapSizeSet(t, (Even(wnew)-1) * xnew, (Even(znew)-1) * ynew))
    return fFalse;
  TX(*t) = xnew; TY(*t) = ynew; TZ(*t) = znew; TW(*t) = wnew;
  return fTrue;
}


// Return whether a pixel in a 3D bitmap is on. Pixels not on the bitmap are
// considered off pixels.

bit Tg(constcubemap t, int x, int y, int z)
{
  return FLegalCube(t, x, y, z) ? TG(t, x, y, z) : fOff;
}


// Set all pixels within a 3D rectangle in a 3D bitmap.

void CubeBlock(cubemap t, int x1, int y1, int z1, int x2, int y2, int z2,
  bit o)
{
  int z;

  SortN(&z1, &z2);
  for (z = z1; z <= z2; z++)
    Block(t, Tx(t, x1, z), Ty(t, y1, z), Tx(t, x2, z), Ty(t, y2, z), o);
}


// Set all pixels within a 3D rectangle in a color 3D bitmap.

void CubeBlockC(colmap c, int x1, int y1, int z1, int x2, int y2, int z2,
  KV kv)
{
  int z;

  SortN(&z1, &z2);
  for (z = z1; z <= z2; z++)
    BlockC(c, Tx(c, x1, z), Ty(c, y1, z), Tx(c, x2, z), Ty(c, y2, z), kv);
}


// Given a coordinates in a 3D bitmap, and an indicator for an axis to focus
// on, return that value transformed appropriately. Used by CubeFlip().

int CubeFlipCore(cubemap t, int x, int y, int z, char ch)
{
  switch (ch) {
  case 'x': return x;
  case 'y': return y;
  case 'z': return z;
  case 'X': return TX(t) - x - 1;
  case 'Y': return TY(t) - y - 1;
  case 'Z': return TZ(t) - z - 1;
  }
  Assert(fFalse);
  return 0;
}


// Transform a 3D bitmap by remapping each of its axes as specified.

void CubeFlip(cubemap *pt, colmap *pc, CONST char *rgch)
{
  cubemap t, t2;
  colmap c2;
  int x, y, z, x2, y2, z2;
  flag fColor = (pt == NULL);

  // Parameter rgch is three characters long, where each indicates how to
  // transform the x, y, and z axes of the 3D bitmap.
  if ((!FBetween(rgch[0], 'X', 'Z') && !FBetween(rgch[0], 'x', 'z')) ||
    (!FBetween(rgch[1], 'X', 'Z') && !FBetween(rgch[1], 'x', 'z')) ||
    (!FBetween(rgch[2], 'X', 'Z') && !FBetween(rgch[2], 'x', 'z')) ||
    rgch[3] != chNull) {
    PrintSz_E("Bad CubeFlip axis or string not 3 characters long.\n");
    return;
  }
  t = !fColor ? *pt : *pc;

  // Allocate a new 3D bitmap to contain the transformed pixels.
  x = CubeFlipCore(t, TX(t), TY(t), TZ(t), ChUncap(rgch[0]));
  y = CubeFlipCore(t, TX(t), TY(t), TZ(t), ChUncap(rgch[1]));
  z = CubeFlipCore(t, TX(t), TY(t), TZ(t), ChUncap(rgch[2]));
  if (!fColor) {
    if (!FAllocateCube(&t2, x, y, z, TW(t)))
      return;
  } else {
    if (!FAllocateCubeC(&c2, x, y, z, TW(t)))
      return;
  }

  // Copy each pixel to its new coordinate.
  if (!fColor)
    BitmapOff(t2);
  for (z = 0; z < TZ(t); z++)
    for (y = 0; y < TY(t); y++)
      for (x = 0; x < TX(t); x++) {
        x2 = CubeFlipCore(t, x, y, z, rgch[0]);
        y2 = CubeFlipCore(t, x, y, z, rgch[1]);
        z2 = CubeFlipCore(t, x, y, z, rgch[2]);
        if (!fColor) {
          if (TG(t, x, y, z))
            T1(t2, x2, y2, z2);
        } else
          CP3(c2, x2, y2, z2, CG3(t, x, y, z));
      }

  if (!fColor) {
    DeallocateCube(t);
    *pt = t2;
  } else {
    DeallocateColmap(t);
    *pc = c2;
  }
}


// Flip or rotate a 3D bitmap across or around a given axis. Called from the
// 3D Bitmap dialog and the Flip3D operation.

void CubeFlip2(cubemap *t, colmap *c, int nAxis, int nOp)
{
  CONST char *pch;

  Assert(FBetween(nAxis, 0, 2) && FBetween(nOp, 0, 3));
  switch (nAxis*4 + nOp) {
  case 0:  pch = "Xyz"; break; // X Flip
  case 1:  pch = "xzY"; break; // X Rotate right
  case 2:  pch = "xZy"; break; // X Rotate left
  case 3:  pch = "xYZ"; break; // X Rotate across
  case 4:  pch = "xYz"; break; // Y Flip
  case 5:  pch = "zyX"; break; // Y Rotate right
  case 6:  pch = "Zyx"; break; // Y Rotate left
  case 7:  pch = "XyZ"; break; // Y Rotate across
  case 8:  pch = "xyZ"; break; // Z Flip
  case 9:  pch = "Yxz"; break; // Z Rotate right
  case 10: pch = "yXz"; break; // Z Rotate left
  case 11: pch = "XYz"; break; // Z Rotate across
  default: Assert(fFalse);
  }
  CubeFlip(t, c, pch);
}


#define FLegalFillCube(b, x, y, z, o) \
  (FLegalCube(t, x, y, z) && TG(t, x, y, z) != (o))

// Flood an irregular volume in a 3D bitmap with on or off pixels.

flag FCubeFill(cubemap t, int x, int y, int z, bit o)
{
  PT *rgpt;
  long ipt = 0;
  int xnew, ynew, znew, d;

  if (!FLegalFillCube(t, x, y, z, o))
    return fTrue;
  if (F64K(t))
    return fFalse;
  rgpt = RgAllocate(TX(t)*TY(t)*TZ(t), PT);
  if (rgpt == NULL)
    return fFalse;
LSet:
  TP(t, x, y, z, o);
LNext:
  for (d = 0; d < DIRS3; d++) {
    xnew = x + xoff3[d]; ynew = y + yoff3[d]; znew = z + zoff3[d];
    if (FLegalFillCube(t, xnew, ynew, znew, o)) {
      FillPush(Tx(t, x, z), Ty(t, y, z));
      x = xnew; y = ynew; z = znew;
      goto LSet;
    }
  }
  if (ipt > 0) {
    FillPop(x, y);
    z = TzM(t, x, y); x = TxM(t, x); y = TyM(t, y);
    goto LNext;
  }
  DeallocateP(rgpt);
  return fTrue;
}


/*
******************************************************************************
** 3D Bitmap File Routines
******************************************************************************
*/

// Load a Daedalus 3D bitmap from a file, into a new 3D bitmap.

flag FReadCube(cubemap *pt, FILE *file, int w)
{
  cubemap t;
  int x, y, z, xT, nCur;
  char ch, ch2;

  ch = getbyte(); ch2 = getbyte();
  if (ch != 'D' || ch2 != '3') {
    PrintSz_W("This file does not look like a Daedalus 3D bitmap.\n");
    return fFalse;
  }
  ch2 = getbyte();
  fscanf(file, "%d%d%d", &x, &y, &z);
  if (!FAllocateCube(&t, x, y, z, w))
    return fFalse;
  if (*pt != NULL)
    DeallocateCube(*pt);
  *pt = t;
  BitmapOff(t);
  skipcrlf(); skipcrlf();
  if (ch2 == 'C') {

    // Supercompressed: 1 character = 6 pixels, plus run length encoding.
    for (z = 0; z < TZ(t); z++) {
      for (y = 0; y < TY(t); y++) {
        x = 0;
        loop {
          ch = getbyte();
          if (ch < ' ') {
            skiplf();
            break;
          }
          if (ch < 'a') {
            // This character encodes 6 pixels.
            nCur = (int)(ch - '!');
            for (xT = 5; xT >= 0; xT--) {
              if (FOdd(nCur))
                T1(t, x + xT, y, z);
              nCur >>= 1;
            }
            x += 6;
          } else {
            // Check for the duplicate row marker '|'.
            if (ch == '|') {
              Assert(x == 0 && y > 0);
              for (nCur = 0, ch = getbyte(); FDigitCh(ch); ch = getbyte())
                nCur = nCur * 10 + (ch - '0');
              loop {
                BlockMove(t, t, Tx(t, 0, z), Ty(t, y-1, z),
                  Tx(t, TX(t)-1, z), Ty(t, y-1, z), Tx(t, 0, z), Ty(t, y, z));
                if (--nCur <= 0)
                  break;
                y++;
              }
              skipcrlf();
              break;
            }
            // This character contains some multiple of 6 on or off pixels.
            if (ch >= 'n') {
              xT = Tx(t, x, z);
              LineX(t, xT, xT + (ch - 'n' + 2) * 6 - 1, Ty(t, y, z), fOn);
              x += (ch - 'n' + 2) * 6;
            } else
              x += (ch - 'a' + 2) * 6;
          }
        }
      }
      skipcrlf();
    }
  } else if (ch2 == 'B') {

    // Compressed: 1 character = 4 pixels.
    for (z = 0; z < TZ(t); z++) {
      for (y = 0; y < TY(t); y++) {
        x = 0;
        loop {
          ch = getbyte();
          if (ch < ' ') {
            skiplf();
            break;
          }
          nCur = NHex(ch);
          for (xT = 3; xT >= 0; xT--) {
            if (x + xT < TX(t) && FOdd(nCur))
              T1(t, x + xT, y, z);
            nCur >>= 1;
          }
          x += 4;
        }
      }
      skipcrlf();
    }
  } else {

    // Not compressed: 1 character = 1 pixel.
    for (z = 0; z < TZ(t); z++) {
      for (y = 0; y < TY(t); y++) {
        x = 0;
        loop {
          ch = getbyte();
          if (ch < ' ') {
            skiplf();
            break;
          }
          if (ch >= chOn)
            T1(t, x, y, z);
          x++;
        }
      }
      skipcrlf();
    }
  }
  return fTrue;
}


// Save a 3D bitmap to a Daedalus 3D bitmap file.

void WriteCube(constcubemap t, FILE *file, int nCompress, flag fClip)
{
  int x, y, z, xT, xmax, nCur, nSav, cn, cRow = 0;
  flag fBlank;
  byte bT;

  fprintf(file, "D3%c\n%d %d %d\n\n", '@' + nCompress, TX(t), TY(t), TZ(t));
  if (nCompress >= 3) {

    // Supercompressed: 1 character = 6 pixels, plus run length encoding.
    for (z = 0; z < TZ(t); z++) {
      for (y = 0; y <= TY(t); y++) {

        // Check if current row same as previous. If so write '|' dup marker.
        if (y > 0) {
          if (y < TY(t)) {
            for (x = 0; x < TX(t); x++)
              if (TG(t, x, y, z) != TG(t, x, y-1, z))
                goto LNormal;
            cRow++;
            continue;
          }
LNormal:
          // Output number of times row was duplicated
          if (cRow > 0) {
            if (!fBlank || cRow > 1) {
              putbyte('|');
              if (cRow > 1)
                fprintf(file, "%d", cRow);
            }
            putbyte('\n');
            cRow = 0;
          }
        }
        if (y >= TY(t))
          continue;

        // Output a unique row.
        xmax = TX(t) - 1;
        if (fClip)
          while (xmax >= 0 && !TG(t, xmax, y, z))
            xmax--;
        fBlank = (xmax < 0);
        cn = 0;
        for (x = 0; x <= xmax + 6; x += 6) {
          // Compute a number encoding the next 6 pixels in this row.
          nCur = 0;
          for (xT = 0; xT < 6; xT++)
            nCur = (nCur << 1) + (x + xT < TX(t) && TG(t, x + xT, y, z));
          // Check for whether ready to write next character/run to file.
          if (x > 0 && (nCur != nSav ||
            (nCur != 0 && nCur != 63) || cn >= 14 || x > xmax)) {
            if (cn <= 1)
              putbyte('!' + nSav);
            else {
              Assert(nSav == 0 || nSav == 63);
              bT = nSav == 0 ? 'a' - 2 + cn : 'n' - 2 + cn;
              putbyte(bT);
            }
            cn = 0;
          }
          nSav = nCur;
          cn++;
        }
        putbyte('\n');
      }
      putbyte('\n');
    }
  } else if (nCompress == 2) {

    // Compressed: 1 character = 4 pixels.
    for (z = 0; z < TZ(t); z++) {
      for (y = 0; y < TY(t); y++) {
        xmax = TX(t) - 1;
        if (fClip)
          while (xmax >= 0 && !TG(t, xmax, y, z))
            xmax--;
        for (x = 0; x <= xmax; x += 4) {
          nCur = 0;
          for (xT = 0; xT < 4; xT++)
            nCur = (nCur << 1) + (x + xT < TX(t) && TG(t, x + xT, y, z));
          putbyte(ChHex(nCur));
        }
        putbyte('\n');
      }
      putbyte('\n');
    }
  } else {

    // Not compressed: 1 character = 1 pixel.
    for (z = 0; z < TZ(t); z++) {
      for (y = 0; y < TY(t); y++) {
        xmax = TX(t) - 1;
        if (fClip)
          while (xmax >= 0 && !TG(t, xmax, y, z))
            xmax--;
        for (x = 0; x <= xmax; x++)
          putbyte(TG(t, x, y, z) ? chOn : chOff);
        putbyte('\n');
      }
      putbyte('\n');
    }
  }
}

/* threed.cpp */
